from torch import nn
import torch.nn.functional as F

class DistillationLoss(nn.Module):
    def __init__(self, T, weight=0.95):
        super().__init__()
        self.T=T
        self.kl_div = nn.KLDivLoss()
        self.alpha = T*T * 2.0 * weight

    _key_map = {'pred_student': 'pred.pred', 'pred_teacher': 'pred.pred_teacher'}
    def forward(self, pred_student, pred_teacher):
        return self.kl_div(F.log_softmax(pred_student/self.T), F.softmax(pred_teacher/self.T)) * self.alpha


from .base import BaseBucket
from .fixed import FixedBucket, FixedCropBucket
from .long_edge import LongEdgeBucket
from .ratio import RatioBucket, SizeBucket, RatioSizeBucket
from .contrastive import PosNegBucket, TripletBucket


import torch
from torch import nn

class CharbonnierLoss(nn.Module):
    """Charbonnier Loss (L1)"""

    def __init__(self, size_average=True, eps=1e-3):
        super(CharbonnierLoss, self).__init__()
        self.size_average = size_average
        self.eps = eps

    def forward(self, x, y):
        diff = x - y
        loss = torch.sqrt((diff * diff) + (self.eps*self.eps))
        if self.size_average:
            loss = torch.mean(loss)
        return loss




def disable_hf_loggers(is_local_main_process=True):
    if is_local_main_process:
        try:
            import transformers

            transformers.utils.logging.set_verbosity_warning()
        except:
            pass
        try:
            import diffusers

            diffusers.utils.logging.set_verbosity_warning()
        except:
            pass
    else:
        try:
            import transformers

            transformers.utils.logging.set_verbosity_error()
        except:
            pass
        try:
            import diffusers

            diffusers.utils.logging.set_verbosity_error()
        except:
            pass



from torch import nn
import torch.nn.functional as F
import torch

class PseudoLabelLoss(nn.Module):
    def __init__(self, threshold=0.7):
        super().__init__()
        self.threshold=threshold

    _key_map = {'pred': 'pred.pred', 'pred_label': 'pred.pred_teacher'}
    def forward(self, pred, pred_label):
        pseudo_label = torch.softmax(pred_label.detach(), dim=-1)
        max_probs, pseudo_targets_u = torch.max(pseudo_label, dim=-1)
        mask = max_probs.ge(self.threshold).float()

        loss_per_sample = F.cross_entropy(pred, pseudo_targets_u, reduction='none')
        return loss_per_sample * mask


from torch import nn


def change_num_classes(model, num_classes):
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    return model


def load_resnet(num_classes=10):
    import torchvision
    model = torchvision.models.resnet18(weights=torchvision.models.ResNet18_Weights.DEFAULT)
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    return model



from .metrics import MetricContainer
import torch

class TorchMetricsContainer(MetricContainer):

    def update(self, pred, target):
        args, kwargs = self.key_mapper(pred=pred, target=target)
        self.metric.update(*args, **kwargs)

    def finish(self, gather, is_local_main_process):
        total = torch.tensor([self.metric.total]).float()
        metric_all = torch.tensor([self.metric.compute()]).float()

        metric_all = gather(metric_all)
        total = gather(total)
        return metric_all*total/total.sum()


import time
from functools import partial

import torch
from torch.nn import MSELoss

from rainbowneko.ckpt_manager import ckpt_saver
from rainbowneko.parser import neko_cfg
from rainbowneko.train.loggers import CLILogger
from rainbowneko.utils import OneCycleLR

time_format="%Y-%m-%d-%H-%M-%S"

@neko_cfg
def make_cfg():
    return dict(
        exp_dir=f'exps/{time.strftime(time_format)}',
        mixed_precision=None,
        allow_tf32=False,
        seed=114514,

        ckpt_saver=dict(
            model=ckpt_saver()
        ),

        train=dict(
            train_steps=1000,
            train_epochs=None,  # Choose one of [train_steps, train_epochs]
            gradient_accumulation_steps=1,
            workers=4,
            max_grad_norm=1.0,
            set_grads_to_none=False,
            retain_graph=False,
            save_step=100,

            resume=None,

            loss=MSELoss(),
            optimizer=partial(torch.optim.AdamW, weight_decay=1e-3),
            scale_lr=True,  # auto scale lr with total batch size
            scheduler=OneCycleLR(
                _partial_=True,
                warmup_steps=0.2,
            ),

            metric=None,
        ),

        logger=[
            partial(CLILogger, out_path='train.log', log_step=20),
        ],

        model=dict(
            name='model',

            enable_xformers=True,
            gradient_checkpointing=True,
            force_cast_precision=False,
            ema=None,

            wrapper=None,
        ),

        evaluator=None,
    )



import torch
import math
import torch.nn.functional as F
from torch.autograd import Variable

class MSSSIMLoss(torch.nn.Module):
    def __init__(self, size_average=True, max_val=1):
        super(MSSSIMLoss, self).__init__()
        self.size_average = size_average
        self.max_val = max_val
        self.weight = torch.tensor([0.0448, 0.2856, 0.3001, 0.2363, 0.1333])

    def gaussian(self, window_size, sigma):
        gauss = torch.Tensor([math.exp(-(x - window_size//2)**2/float(2*sigma**2)) for x in range(window_size)])
        return gauss/gauss.sum()

    def create_window(self, window_size, sigma, channel):
        _1D_window = self.gaussian(window_size, sigma).unsqueeze(1)
        _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
        window = Variable(_2D_window.expand(channel, 1, window_size, window_size)).contiguous()
        return window

    def _ssim(self, img1, img2, size_average = True):
        _, c, w, h = img1.size()
        window_size = min(w, h, 11)
        sigma = 1.5 * window_size / 11
        window = self.create_window(window_size, sigma, c).to(img1.device)
        mu1 = F.conv2d(img1, window, padding=window_size//2, groups=c)
        mu2 = F.conv2d(img2, window, padding=window_size//2, groups=c)

        mu1_sq = mu1.pow(2)
        mu2_sq = mu2.pow(2)
        mu1_mu2 = mu1*mu2

        sigma1_sq = F.conv2d(img1*img1, window, padding=window_size//2, groups=c) - mu1_sq
        sigma2_sq = F.conv2d(img2*img2, window, padding=window_size//2, groups=c) - mu2_sq
        sigma12 = F.conv2d(img1*img2, window, padding=window_size//2, groups=c) - mu1_mu2

        C1 = (0.01 * self.max_val) **2
        C2 = (0.03 * self.max_val) **2
        V1 = 2.0 * sigma12 + C2
        V2 = sigma1_sq + sigma2_sq + C2
        ssim_map = ((2 * mu1_mu2 + C1) * V1) / ((mu1_sq + mu2_sq + C1) * V2)
        mcs_map = V1 / V2
        if size_average:
            return ssim_map.mean(), mcs_map.mean()

    def ms_ssim(self, img1, img2, levels=5):
        self.weight = self.weight.to(img1.device)
        msssim = torch.empty(levels, device=img1.device)
        mcs = torch.empty(levels, device=img1.device)
        for i in range(levels):
            ssim_map, mcs_map = self._ssim(img1, img2)
            msssim[i] = ssim_map
            mcs[i] = mcs_map
            filtered_im1 = F.avg_pool2d(img1, kernel_size=2, stride=2)
            filtered_im2 = F.avg_pool2d(img2, kernel_size=2, stride=2)
            img1 = filtered_im1
            img2 = filtered_im2

        value = (torch.prod(mcs[0:levels-1] ** self.weight[0:levels-1])*
                                    (msssim[levels-1] ** self.weight[levels-1]))
        return value


    def forward(self, img1, img2):
        img1 = torch.relu(img1)
        img2 = torch.relu(img2)
        return 1-self.ms_ssim(img1, img2)


from typing import Tuple, Union

import numpy as np

from .base import BaseBucket
from ..handler import AutoSizeHandler


class PosNegBucket(BaseBucket):
    can_shuffle = False
    handler = AutoSizeHandler(mode='full')

    def __init__(self, target_size: Union[Tuple[int, int], int] = 512, pos_rate=0.5, **kwargs):
        self.target_size = (target_size, target_size) if isinstance(target_size, int) else target_size
        self.pos_rate = pos_rate

    def build(self, bs: int, world_size: int, source: 'DataSource'):
        self.bs = bs  # bs per device
        self.world_size = world_size
        self.source = source

        self.cls_group = {}  # {cls: idx}
        for i, (data, source) in enumerate(self.source):
            cls_name = data['label']
            if cls_name not in self.cls_group:
                self.cls_group[cls_name] = []
            self.cls_group[cls_name].append(i)
        self.cls_group = {k: np.array(v) for k, v in self.cls_group.items()}
        self.cls_names = list(self.cls_group.keys())

    def rest(self, epoch):
        self.rs = np.random.RandomState(42 + epoch)

        # build bucket
        bucket = []
        pos_len = int(self.bs * self.pos_rate)
        neg_len = self.bs - pos_len
        for name, group in self.cls_group.items():
            rest = len(group) % pos_len
            if rest > 0:
                group = np.hstack([group, self.rs.choice(group, pos_len - rest)])

            group = group.reshape(-1, pos_len)
            neg = np.hstack([g_neg for name_neg, g_neg in self.cls_group.items() if name_neg != name])
            group_neg = self.rs.choice(neg, (group.shape[0], neg_len), replace=False)
            group = np.hstack([group, group_neg])
            bucket.append(group)

        bucket = np.vstack(bucket)  # [N,bs]
        rest = len(bucket) % self.world_size
        if rest > 0:
            idx = self.rs.random_integers(0, len(bucket) - 1, self.world_size - rest)
            bucket = np.vstack([bucket, bucket[idx]])
        bucket = bucket.reshape(-1, self.world_size, self.bs).transpose(0, 2, 1)

        self.img_idxs = bucket.flatten()
        self.data_len = len(self.img_idxs)

    def __getitem__(self, idx) -> Tuple[Tuple[str, 'DataSource'], Tuple[int, int]]:
        file_idx = self.img_idxs[idx]
        datas = self.source[file_idx]
        datas['image_size'] = self.target_size
        return datas

    def __len__(self):
        return self.data_len


class TripletBucket(PosNegBucket):
    def rest(self, epoch):
        assert self.bs % 3 == 0, 'batch size of TripletBucket must be a multiple of 3.'
        self.rs = np.random.RandomState(42 + epoch)

        # build bucket
        bucket = []
        pos_len = 2 * self.bs // 3
        neg_len = self.bs - pos_len
        for name, group in self.cls_group.items():
            rest = len(group) % pos_len
            if rest > 0:
                group = np.hstack([group, self.rs.choice(group, pos_len - rest)])

            group = group.reshape(-1, pos_len)
            neg = np.hstack([g_neg for name_neg, g_neg in self.cls_group.items() if name_neg != name])
            group_neg = self.rs.choice(neg, (group.shape[0], neg_len), replace=False)
            group = np.hstack([group, group_neg])
            bucket.append(group)

        bucket = np.vstack(bucket)  # [N,bs]
        rest = len(bucket) % self.world_size
        if rest > 0:
            idx = self.rs.random_integers(0, len(bucket) - 1, self.world_size - rest)
            bucket = np.vstack([bucket, bucket[idx]])

        # shuffle triplets
        bucket = bucket.reshape(-1, 3, self.bs // 3).transpose(0, 2, 1)
        self.rs.shuffle(bucket)
        bucket = bucket.transpose(0, 2, 1).reshape(-1, self.bs)

        bucket = bucket.reshape(-1, self.world_size, self.bs).transpose(0, 2, 1)

        self.img_idxs = bucket.flatten()
        self.data_len = len(self.img_idxs)

    def __getitem__(self, idx) -> Tuple[Tuple[str, 'DataSource'], Tuple[int, int]]:
        file_idx = self.img_idxs[idx]
        datas = self.source[file_idx]
        datas['image_size'] = self.target_size
        return datas



from .base import DataHandler
import torch
from torch.nn import functional as F

class MixUPHandler(DataHandler):
    def __init__(self, num_classes: int, alpha: float = 1.0, ratio: float = 1.0,
                 key_map_in=('image -> image', 'label -> label'), key_map_out=('image -> image', 'label -> label')):
        super().__init__(key_map_in, key_map_out)
        self.num_classes = num_classes
        self.alpha = alpha
        self.ratio = ratio
        self._dist = torch.distributions.Beta(torch.tensor([alpha]), torch.tensor([alpha]))

    def handle(self, image, label):
        if label.dim() == 1:
            label = F.one_hot(label, num_classes=self.num_classes)

        label = label.float()
        N_mix = int(image.shape[0] * self.ratio)
        lam = self._dist.sample((N_mix,))
        lam_i = lam.view(-1, *[1] * (image.dim() - 1))
        lam_t = lam.view(-1, 1)

        image[:N_mix] = image[:N_mix].roll(1, 0).mul_(1.0 - lam_i).add_(image[:N_mix].mul(lam_i))
        label[:N_mix] = label[:N_mix].roll(1, 0).mul_(1.0 - lam_t).add_(label[:N_mix].mul(lam_t))

        return {'image': image, 'label': label}


"""
ckpt_pkl.py
====================
    :Name:        save model with torch
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     8/04/2023
    :Licence:     MIT
"""

from typing import Dict

from rainbowneko.models.ema import ModelEMA
from rainbowneko.models.plugin import PluginGroup
from torch import nn

from .ckpt import NekoModelSaver, NekoPluginSaver


class NekoModelModuleSaver(NekoModelSaver):
    def _get_modules(self, model):
        assert len(self.target_module) == 1, "Only one target module is supported for NekoModelModuleSaver"

        item = self.target_module[0]
        base_model = model if item == '' else eval(f"model.{item}")
        return base_model

    def save_to(self, name, model: nn.Module, plugin_groups: Dict[str, PluginGroup], model_ema: ModelEMA = None, exclude_key=None,
                name_template=None):
        sd_model = {"base": self._get_modules(model)}
        if model_ema is not None:
            sd_model["base_ema"] = self._get_modules(model_ema.model)

        if name_template is not None:
            name = name_template.format(name)
        self.save(sd_model, name, prefix=self.prefix)


class NekoPluginModuleSaver(NekoPluginSaver):

    def save_to(self, name, model: nn.Module, plugin_groups: Dict[str, PluginGroup], model_ema: ModelEMA = None, exclude_key=None,
                name_template=None):
        plugin_name = self.target_plugin[0]
        plugin = plugin_groups[plugin_name]
        sd_plugin = {"plugin": plugin}
        if model_ema is not None:
            sd_plugin["plugin_ema"] = plugin.from_model(model_ema.model)

        if name_template is not None:
            name = name_template.format(name)
        self.save(sd_plugin, name, prefix=self.prefix)



from functools import partial


class RecursivePartial(partial):
    __slots__ = "func", "args", "keywords", "add_sub"

    def __new__(cls, func, /, *args, **keywords):
        if not callable(func):
            raise TypeError("the first argument must be callable")

        if hasattr(func, "func"):
            args = func.args + args
            keywords = {**func.keywords, **keywords}
            func = func.func

        self = super(RecursivePartial, cls).__new__(cls, func)

        self.func = func
        self.args = [args]
        self.keywords = [keywords]
        self.add_sub = False
        return self

    def __call__(self, /, *args, **keywords):
        if self.add_sub:
            self.args.append(args)
            self.keywords.append(keywords)
            self.add_sub = False
            return self
        else:
            keywords = {**self.keywords[0], **keywords}
            res = self.func(*self.args[0], *args, **keywords)
            for args_i, kwargs_i in zip(self.args[1:], self.keywords[1:]):
                if len(args_i) == 2 and args_i[0] is getattr:
                    res = getattr(res, args_i[1])
                else:
                    res = res(*args_i, **kwargs_i)
            return res

    def __repr__(self):
        fdict = lambda d: ', '.join(f'{k}={v}' for k, v in d.items())
        flist = lambda l: ', '.join(f'{k}' for k in l) + ', ' if len(l) > 0 else ''

        pstr = f'{self.__class__.__name__}[{self.func}({flist(self.args[0])}{fdict(self.keywords[0])})'
        for args_i, kwargs_i in zip(self.args[1:], self.keywords[1:]):
            if len(args_i) == 2 and args_i[0] is getattr:
                pstr += f'.{args_i[1]}'
            else:
                pstr += f'({flist(args_i)}{fdict(kwargs_i)})'
        return pstr+']'



import importlib.resources as pkg_resources
import os.path
import shutil
from contextlib import contextmanager


@contextmanager
def _ctx_package_dir(package_name):
    # don't use the f*cking except here, method `files`'s existence can be checked with `hasattr` easily
    if hasattr(pkg_resources, 'files'):
        # 对于Python 3.9及以上版本，使用files函数
        yield str(pkg_resources.files(package_name).joinpath(''))
    else:
        # 对于Python 3.7和3.8，使用path函数
        with pkg_resources.path(package_name, '') as package_dir:
            yield package_dir


def copy_package_data(package_name, package_subdir, destination):
    """
    将指定包内的数据目录复制到目标路径。
    :param package_name: 包的完整名称，例如 'my_package.data'
    :param package_subdir: Sub directory inside the package.
    :param destination: 目标文件夹路径，例如 '/path/to/destination'
    """
    with _ctx_package_dir(package_name) as package_dir:
        src_dir = os.path.join(package_dir, package_subdir)
        shutil.copytree(src_dir, destination, dirs_exist_ok=True)


def main():
    copy_package_data('rainbowneko', 'cfgs', './cfgs')



import hashlib
import math
import random
import re
from itertools import cycle, islice
from pathlib import Path
from typing import Tuple, List, Any, Dict, Union

import torch
from omegaconf import OmegaConf

from .img_size_tool import types_support

size_mul = {'K': 1 << 10, 'M': 1 << 20, 'G': 1 << 30, 'T': 1 << 40}
size_key = 'TGMK'


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def low_rank_approximate(weight, rank, clamp_quantile=0.99):
    if len(weight.shape) == 4:  # conv
        weight = weight.flatten(1)
        out_ch, in_ch, k1, k2 = weight.shape

    U, S, Vh = torch.linalg.svd(weight)
    U = U[:, :rank]
    S = S[:rank]
    U = U @ torch.diag(S)

    Vh = Vh[:rank, :]

    dist = torch.cat([U.flatten(), Vh.flatten()])
    hi_val = torch.quantile(dist, clamp_quantile)
    low_val = -hi_val

    U = U.clamp(low_val, hi_val)
    Vh = Vh.clamp(low_val, hi_val)

    if len(weight.shape) == 4:
        # U is (out_channels, rank) with 1x1 conv.
        U = U.reshape(U.shape[0], U.shape[1], 1, 1)
        # V is (rank, in_channels * kernel_size1 * kernel_size2)
        Vh = Vh.reshape(Vh.shape[0], in_ch, k1, k2)
    return U, Vh


def get_cfg_range(cfg_text: str):
    dy_cfg_f = 'ln'
    if cfg_text.find(':') != -1:
        cfg_text, dy_cfg_f = cfg_text.split(':')

    if cfg_text.find('-') != -1:
        l, h = cfg_text.split('-')
        return float(l), float(h), dy_cfg_f
    else:
        return float(cfg_text), float(cfg_text), dy_cfg_f


def to_validate_file(name):
    rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
    new_title = re.sub(rstr, "_", name)  # 替换为下划线
    return new_title[:50]


def make_mask(start, end, length):
    mask = torch.zeros(length)
    mask[int(length * start):int(length * end)] = 1
    return mask.bool()


def is_image_file(file: Union[str, Path]):
    return Path(file).suffix[1:].lower() in types_support


def factorization(dimension: int, factor: int = -1) -> Tuple[int, int]:
    find_one = lambda x: len(x) - (x.rfind('1') + 1)
    dim_bin = bin(dimension)
    num = find_one(dim_bin)
    f_max = (len(dim_bin) - 3) >> 1 if factor < 0 else find_one(bin(factor))
    num = min(num, f_max)
    return dimension >> num, 1 << num


def isinstance_list(obj, cls_list):
    for cls in cls_list:
        if isinstance(obj, cls):
            return True
    return False


def net_path_join(*args):
    return '.'.join(args).strip('.').replace('..', '.')


def mgcd(*args):
    g = args[0]
    for s in args[1:]:
        g = math.gcd(g, s)
    return g


def size_to_int(size):
    return int(size[:-3]) * size_mul[size[-3]]


def int_to_size(size):
    for i, k in zip(range(40, 0, -10), size_key):
        if size >= 1 << i:
            return f'{size >> i}{k}iB'


def prepare_seed(seeds: List[int], device='cuda'):
    return [torch.Generator(device=device).manual_seed(s or random.randint(0, 1 << 30)) for s in seeds]


def hash_str(data):
    return hashlib.sha256(data.encode('utf-8')).hexdigest()


def format_number(num):
    if num >= 1e9:
        return f'{num / 1e9:.1f}B'
    elif num >= 1e6:
        return f'{num / 1e6:.1f}M'
    elif num >= 1e3:
        return f'{num / 1e3:.1f}K'
    else:
        return str(num)


def is_list(v):
    return OmegaConf.is_list(v) or isinstance(v, list)


def is_dict(v):
    return OmegaConf.is_dict(v) or isinstance(v, dict)


def addto_dictlist(dict_list: Dict[str, List], data: Dict[str, Any], v_proc=None):
    for k, v in data.items():
        if k not in dict_list:
            dict_list[k] = []
        if v_proc is not None:
            dict_list[k].append(v_proc(v))
        else:
            dict_list[k].append(v)
    return dict_list


def set_list_value(lst, index, value, default=None):
    if index >= len(lst):
        lst.extend([default] * (index - len(lst) + 1))
    lst[index] = value


def dict_parse_list(data: Dict[str, Any]):
    if isinstance(data, dict):
        if len(data) > 0 and isinstance(next(iter(data)), int):
            return [dict_parse_list(data[i]) for i in range(len(data))]
        else:
            return {k: dict_parse_list(v) for k, v in data.items()}
    else:
        return data

def dict_list_copy(obj):
    if isinstance(obj, dict):
        return {key: dict_list_copy(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [dict_list_copy(item) for item in obj]
    else:
        return obj

def remove_empty_dict_list(data):
    if isinstance(data, dict):
        cleaned = {k: remove_empty_dict_list(v) for k, v in data.items()}
        return {k: v for k, v in cleaned.items() if v not in ({}, [])}
    elif isinstance(data, list):
        cleaned = [remove_empty_dict_list(v) for v in data]
        return [v for v in cleaned if v not in ({}, [])]
    return data

def dict_merge(dict_base, dict_override):
    """
    recursive merge dicts
    """
    merged = dict_base.copy()
    for key, value in dict_override.items():
        if key in merged:
            if isinstance(merged[key], dict) and isinstance(value, dict):
                merged[key] = dict_merge(merged[key], value)
            else:
                merged[key] = value
        else:
            merged[key] = value
    return merged


def repeat_list(lst, length):
    return list(islice(cycle(lst), length))



from cfgs.py.train.classify import multi_class
from rainbowneko.data import BaseDataset
from rainbowneko.data.handler import MixUPHandler, HandlerChain
from rainbowneko.parser import neko_cfg
from rainbowneko.train.loss import LossContainer, SoftCELoss

num_classes = 10
multi_class.num_classes = num_classes

@neko_cfg
def make_cfg():
    return dict(
        _base_=[multi_class],

        train=dict(
            loss=LossContainer(loss=SoftCELoss()),
            metrics=None,
        ),

        data_train=dict(
            dataset1=BaseDataset(
                batch_handler=HandlerChain(
                    mixup=MixUPHandler(num_classes=num_classes)
                )
            )
        ),
    )



from torch import nn
from rainbowneko.utils import KeyMapper

class LossContainer(nn.Module):
    def __init__(self, loss, weight=1.0, key_map=None):
        super().__init__()
        self.key_mapper = KeyMapper(loss, key_map)
        self.loss = loss
        self.weight = weight

    def forward(self, pred, inputs):
        args, kwargs = self.key_mapper(pred=pred, inputs=inputs)
        return self.loss(*args, **kwargs) * self.weight

class LossGroup(nn.Module):
    def __init__(self, loss_list):
        super().__init__()
        self.loss_list = loss_list

    def forward(self, pred, target):
        loss = 0
        for loss_item in self.loss_list:
            loss += loss_item(pred, target).squeeze()
        return loss


from cfgs.py.train.classify import multi_class
from rainbowneko.ckpt_manager import ckpt_saver, LocalCkptSource, NekoModelModuleSaver
from rainbowneko.ckpt_manager.format import ONNXFormat
from rainbowneko.parser import neko_cfg


@neko_cfg
def make_cfg():
    return dict(
        _base_=[multi_class],

        ckpt_saver=dict(
            model=ckpt_saver(target_module='model'),
            model_onnx=NekoModelModuleSaver(
                format=ONNXFormat(inputs={'image': (('batch', 1), 3, 32, 32)}),
                source=LocalCkptSource(),
                target_module='model'
            )
        ),

    )



import os
from typing import Dict, Any

from PIL import Image
from loguru import logger

from .base_logger import BaseLogger

class CLILogger(BaseLogger):
    def __init__(self, exp_dir, out_path, log_step=10,
                 img_log_dir=None, img_ext='png', img_quality=95):
        super().__init__(exp_dir, out_path, log_step)
        if exp_dir is not None:  # exp_dir is only available in local main process
            if out_path is not None:
                logger.add(os.path.join(exp_dir, out_path))
            if img_log_dir is not None:
                self.img_log_dir = os.path.join(exp_dir, img_log_dir)
                os.makedirs(self.img_log_dir, exist_ok=True)
            self.img_ext = img_ext
            self.img_quality = img_quality
        else:
            self.disable()

    def enable(self):
        super(CLILogger, self).enable()
        logger.enable("__main__")

    def disable(self):
        super(CLILogger, self).disable()
        logger.disable("__main__")

    def _info(self, info):
        logger.info(info)

    def log_text(self, datas: Dict[str, Any], step: int = 0):
        logger.info(', '.join([f"{os.path.basename(k)} = {v['format'].format(*v['data'])}" for k, v in datas.items()]))

    def log_image(self, imgs: Dict[str, Image.Image], step: int = 0):
        logger.info(f'log {len(imgs)} images')
        for name, img in imgs.items():
            img.save(os.path.join(self.img_log_dir, f'{step}-{name}.{self.img_ext}'), quality=self.img_quality)



"""
lora.py
====================
    :Name:        lora tools
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     10/03/2023
    :Licence:     Apache-2.0
"""

from typing import Union, Tuple

import torch
from rainbowneko.utils.utils import low_rank_approximate
from torch import nn

from ..plugin import PatchPluginBlock, PluginGroup, PatchPluginContainer


class LoraPatchContainer(PatchPluginContainer):
    def forward(self, x, *args, **kwargs):
        weight_ = None
        bias_ = None
        for name in self.plugin_names:
            if weight_ is None:
                weight_ = self[name].get_weight()
            else:
                weight_ = weight_ + self[name].get_weight()

            if bias_ is None:
                bias_ = self[name].get_bias()
            else:
                bias_ = bias_ + self[name].get_bias()

        return self[name].post_forward(x, self._host.weight, weight_, self._host.bias, bias_)


class LoraBlock(PatchPluginBlock):
    container_cls = LoraPatchContainer
    wrapable_classes = (nn.Linear, nn.Conv2d)

    def __init__(self, lora_id: int, host: Union[nn.Linear, nn.Conv2d], rank, dropout=0.1, alpha=1.0, bias=False,
                 alpha_auto_scale=True, parent_block=None, host_name=None, **kwargs):
        super().__init__(f'lora_block_{lora_id}', host, parent_block=parent_block, host_name=host_name)

        self.bias = bias

        host = self.host()
        if isinstance(host, nn.Linear):
            self.host_type = 'linear'
            self.layer = self.LinearLayer(host, rank, bias, self)
        elif isinstance(host, nn.Conv2d):
            self.host_type = 'conv'
            self.layer = self.Conv2dLayer(host, rank, bias, self)
        else:
            raise NotImplementedError(f'No lora for {type(host)}')
        self.dropout = nn.Dropout(dropout)

        self.rank = self.layer.rank
        self.register_buffer('alpha', torch.tensor(alpha / self.rank if alpha_auto_scale else alpha))

    def extra_repr(self) -> str:
        return f'rank={self.rank}, alpha={self.alpha}, bias={self.bias}'

    def get_weight(self):
        return self.layer.get_weight() * self.alpha

    def get_bias(self):
        bias = self.layer.get_bias()
        return bias * self.alpha if bias is not None else None

    def post_forward(self, x, host_weight, weight, host_bias, bias=None):
        if host_bias is not None:
            if bias is None:
                bias = host_bias
            else:
                bias = host_bias + bias
        return self.dropout(self.layer(x, host_weight + weight, bias))

    def init_weights(self, svd_init=False):
        if svd_init:
            host = self.host()
            U, V = low_rank_approximate(host.weight, self.rank)
            self.layer.feed_svd(U, V, host.weight)
        else:
            self.layer.reset_parameters()

    def reparameterization_to_host(self, alpha=None, base_alpha=1.0):
        if alpha is None:
            alpha = self.alpha

        host = self.host()
        re_w, re_b = self.layer.get_collapsed_param()
        host.weight = nn.Parameter(
            host.weight.data * base_alpha + alpha * re_w.to(host.weight.device, dtype=host.weight.dtype)
        )

        if re_b is not None:
            if host.bias is None:
                host.bias = nn.Parameter(re_b.to(host.weight.device, dtype=host.weight.dtype))
            else:
                host.bias = nn.Parameter(
                    host.bias.data * base_alpha + alpha * re_b.to(host.weight.device, dtype=host.weight.dtype))

    class LinearLayer(nn.Module):
        def __init__(self, host, rank, bias, block):
            super().__init__()
            self.rank = rank
            if isinstance(self.rank, float):
                self.rank = max(round(host.out_features * self.rank), 1)

        def feed_svd(self, U, V, weight):
            self.lora_up.weight.data = U.to(device=weight.device, dtype=weight.dtype)
            self.lora_down.weight.data = V.to(device=weight.device, dtype=weight.dtype)

        def get_weight(self) -> torch.Tensor:
            pass

        def get_bias(self) -> torch.Tensor:
            pass

        def forward(self, x, weight, bias=None):
            pass

        def get_collapsed_param(self) -> Tuple[torch.Tensor, torch.Tensor]:
            pass

    class Conv2dLayer(nn.Module):
        def __init__(self, host, rank, bias, block):
            super().__init__()
            self.rank = rank
            if isinstance(self.rank, float):
                self.rank = max(round(host.out_channels * self.rank), 1)

        def feed_svd(self, U, V, weight):
            self.lora_up.weight.data = U.to(device=weight.device, dtype=weight.dtype)
            self.lora_down.weight.data = V.to(device=weight.device, dtype=weight.dtype)

        def get_weight(self) -> torch.Tensor:
            pass

        def get_bias(self) -> torch.Tensor:
            pass

        def forward(self, x, weight, bias=None):
            pass

        def get_collapsed_param(self) -> Tuple[torch.Tensor, torch.Tensor]:
            pass

    @classmethod
    def wrap_layer(cls, lora_id: int, layer: Union[nn.Linear, nn.Conv2d], rank=1, dropout=0.0, alpha=1.0, svd_init=False,
                   bias=False, mask=None, **kwargs):  # -> LoraBlock:
        lora_block = cls(lora_id, layer, rank, dropout, alpha, bias=bias, **kwargs)
        lora_block.init_weights(svd_init)
        return lora_block

    @classmethod
    def wrap_model(cls, lora_id: int, model: nn.Module, **kwargs):  # -> Dict[str, LoraBlock]:
        return super(LoraBlock, cls).wrap_model(lora_id, model, exclude_classes=(LoraBlock,), parent_block=True, **kwargs)

    @staticmethod
    def extract_lora_state(model: nn.Module):
        return {k: v for k, v in model.state_dict().items() if 'lora_block_' in k}

    @staticmethod
    def extract_state_without_lora(model: nn.Module):
        return {k: v for k, v in model.state_dict().items() if 'lora_block_' not in k}

    @staticmethod
    def extract_param_without_lora(model: nn.Module):
        return {k: v for k, v in model.named_parameters() if 'lora_block_' not in k}

    @staticmethod
    def extract_trainable_state_without_lora(model: nn.Module):
        trainable_keys = {k for k, v in model.named_parameters() if ('lora_block_' not in k) and v.requires_grad}
        return {k: v for k, v in model.state_dict().items() if k in trainable_keys}


class LoraGroup(PluginGroup):
    def set_mask(self, batch_mask):
        for item in self.plugin_dict.values():
            item.set_mask(batch_mask)

    def collapse_to_host(self, alpha=None, base_alpha=1.0):
        for item in self.plugin_dict.values():
            item.collapse_to_host(alpha, base_alpha)

    def set_inplace(self, inplace):
        for item in self.plugin_dict.values():
            item.inplace = inplace


def split_state(state_dict):
    sd_base, sd_lora = {}, {}
    for k, v in state_dict.items():
        if 'lora_block_' in k:
            sd_lora[k] = v
        else:
            sd_base[k] = v
    return sd_base, sd_lora



from typing import Dict, Any

import os
import wandb
from PIL import Image

from .base_logger import BaseLogger


class WanDBLogger(BaseLogger):
    def __init__(self, exp_dir, out_path=None, project='rainbow-neko', log_step=10):
        super().__init__(exp_dir, out_path, log_step)
        if exp_dir is not None:  # exp_dir is only available in local main process
            wandb.init(project=project, name=os.path.basename(exp_dir))
            wandb.save(os.path.join(exp_dir, 'cfg.yaml'), base_path=exp_dir)
        else:
            self.writer = None
            self.disable()

    def _info(self, info):
        pass

    def log_text(self, datas: Dict[str, Any], step: int = 0):
        log_dict = {'step': step}
        for k, v in datas.items():
            if len(v['data']) == 1:
                log_dict[k] = v['data'][0]
        wandb.log(log_dict)

    def log_image(self, imgs: Dict[str, Image.Image], step: int = 0):
        wandb.log({next(iter(imgs.keys())): list(imgs.values())}, step=step)



from typing import Tuple, Union

from .base import BaseBucket
from ..handler import AutoSizeHandler


class FixedBucket(BaseBucket):
    handler = AutoSizeHandler()

    def __init__(self, target_size: Union[Tuple[int, int], int] = 512, **kwargs):
        self.target_size = (target_size, target_size) if isinstance(target_size, int) else target_size

    def __getitem__(self, idx) -> Tuple[Tuple[str, 'DataSource'], Tuple[int, int]]:
        datas = self.source[idx]
        datas['image_size'] = self.target_size
        return datas

    def __len__(self):
        return len(self.source)


class FixedCropBucket(FixedBucket):
    handler = AutoSizeHandler(mode='pad')



import glob
import json
import os
from pathlib import Path
from typing import Dict

import yaml
from loguru import logger
from rainbowneko.utils.img_size_tool import types_support


class BaseLabelLoader:
    def __init__(self, path: str):
        self.path = Path(path)

    def _load(self):
        raise NotImplementedError

    def load(self):
        ''' {image.ext: label} '''
        retval = self._load()
        logger.info(f'{len(retval)} record(s) loaded with {self.__class__.__name__}, from path {self.path!r}')
        return retval

    @staticmethod
    def clean_ext(captions: Dict[str, str]):
        ''' image.ext -> image '''

        def rm_ext(path):
            name, ext = os.path.splitext(path)
            if len(ext) > 0 and ext[1:] in types_support:
                return name
            return path

        return {rm_ext(k): v for k, v in captions.items()}


class JsonLabelLoader(BaseLabelLoader):
    def _load(self):
        with open(self.path, 'r', encoding='utf-8') as f:
            return json.loads(f.read())


class YamlLabelLoader(BaseLabelLoader):
    def _load(self):
        with open(self.path, 'r', encoding='utf-8') as f:
            return yaml.load(f.read(), Loader=yaml.FullLoader)


class TXTLabelLoader(BaseLabelLoader):
    def _load(self):
        # get all txt files and their related images
        img_map = {}
        txt_list = []
        for file in self.path.iterdir():
            file: Path
            if file.is_file():
                if file.suffix == '.txt':
                    txt_list.append(file)
                else:
                    img_map[file.stem] = file.name

        # load captions
        captions = {}
        for txt in txt_list:
            with open(txt, 'r', encoding='utf-8') as f:
                captions[img_map[txt.stem]] = f.read().strip()
        return captions


def auto_label_loader(path):
    if os.path.isdir(path):
        json_files = glob.glob(os.path.join(path, '*.json'))
        if json_files:
            return JsonLabelLoader(json_files[0])

        yaml_files = [
            *glob.glob(os.path.join(path, '*.yaml')),
            *glob.glob(os.path.join(path, '*.yml')),
        ]
        if yaml_files:
            return YamlLabelLoader(yaml_files[0])

        txt_files = glob.glob(os.path.join(path, '*.txt'))
        if txt_files:
            return TXTLabelLoader(path)

        raise FileNotFoundError(f'Caption file not found in directory {path!r}.')

    elif os.path.isfile(path):
        _, ext = os.path.splitext(path)
        if ext == '.json':
            return JsonLabelLoader(path)
        elif ext in {'.yaml', '.yml'}:
            return YamlLabelLoader(path)
        else:
            raise FileNotFoundError(f'Unknown caption file {path!r}.')

    else:
        raise FileNotFoundError(f'Unknown caption file type {path!r}.')



import argparse

import hydra
import torch

from rainbowneko.parser import load_config_with_cli, YamlCfgParser
from .workflow import BasicAction
from rainbowneko import _share


class WorkflowRunner:
    def __init__(self, parser: YamlCfgParser, cfgs):
        self.cfgs_raw = cfgs
        cfgs = hydra.utils.instantiate(cfgs)
        self.cfgs = cfgs
        self.parser = parser

        if _share.loggers is None:
            from rainbowneko.train.loggers import CLILogger
            _share.loggers = CLILogger('exps/', None, log_step=1)

        self.actions: BasicAction = cfgs.workflow

    @torch.no_grad()
    def run(self, **states_in):
        states = dict(cfgs=self.cfgs_raw, parser=self.parser, world_size=1, local_rank=0)
        if states_in is not None:
            states.update(states_in)
        states = self.actions(**states)
        return states


def run_workflow():
    parser = argparse.ArgumentParser(description='RainbowNeko Workflow Launcher')
    parser.add_argument('--cfg', type=str, default='')
    args, cfg_args = parser.parse_known_args()

    import sys
    import os
    sys.path.append(os.getcwd())

    parser, conf = load_config_with_cli(args.cfg, args_list=cfg_args)
    runner = WorkflowRunner(parser, conf)
    runner.run()



from functools import partial
from typing import Dict

import torch
from rainbowneko.utils import addto_dictlist, KeyMapper

class BaseMetric:
    def __init__(self):
        pass

    def reset(self):
        raise NotImplementedError

    def update(self, pred, target):
        raise NotImplementedError

    def finish(self, gather, is_local_main_process):
        raise NotImplementedError

    def to(self, device):
        pass


class MetricContainer(BaseMetric):
    def __init__(self, metric, device='cpu', key_map=None):
        super().__init__()
        self.key_mapper = KeyMapper(metric, key_map)
        self.metric = metric
        self.device = device

    def reset(self):
        self.metric_list = []

    def update(self, pred, inputs):
        args, kwargs = self.key_mapper(pred=pred, inputs=inputs)
        v_metric = self.metric(*args, **kwargs)
        self.metric_list.append(v_metric)

    def finish(self, gather, is_local_main_process):
        total = torch.tensor([len(self.metric_list)]).float()
        try:
            metric_all = torch.cat(self.metric_list).mean()
        except:
            metric_all = torch.tensor(self.metric_list).mean()

        metric_all = metric_all.to(self.device)
        total = total.to(self.device)
        if gather is not None:
            metric_all = gather(metric_all)
            total = gather(total)
        return (metric_all*total/total.sum()).sum().item()

    def to(self, device):
        self.device = device
        if hasattr(self.metric, 'to'):
            self.metric.to(device)

class FullMetricContainer(MetricContainer):

    def reset(self):
        self.args_list = []
        self.kwargs_list = {}

    def update(self, pred, target):
        args, kwargs = self.key_mapper(pred=pred, target=target)

        if len(self.args_list) == 0:
            for v in args:
                self.args_list.append([v])
        else:
            for i,v in enumerate(args):
                self.args_list[i].append(v)

        addto_dictlist(self.kwargs_list, kwargs)

    def finish(self, gather, is_local_main_process):
        for i, v in enumerate(self.args_list):
            self.args_list[i] = torch.cat(v)

        for k, v in self.kwargs_list.items():
            self.kwargs_list[k] = torch.cat(v)

        self.args_list = gather(self.args_list)
        self.kwargs_list = gather(self.kwargs_list)

        v_metric = self.metric(*self.args_list, **self.kwargs_list)

        return v_metric.item()


class MetricGroup(BaseMetric):
    def __init__(self, **metric_dict: BaseMetric):
        self.metric_dict = {k: (v() if isinstance(v, partial) else v) for k, v in metric_dict.items()}

    def reset(self):
        for name, metric in self.metric_dict.items():
            metric.reset()

    def update(self, *args, **kwargs):
        for name, metric in self.metric_dict.items():
            metric.update(*args, **kwargs)

    def finish(self, gather, is_local_main_process):
        metric_dict = {}
        for name, metric in self.metric_dict.items():
            metric_dict[name] = metric.finish(gather, is_local_main_process)
        return metric_dict

    def to(self, device):
        for metric in self.metric_dict.values():
            metric.to(device)

    @staticmethod
    def format(metrics_dict, format="{:.2e}", prefix=''):
        if not isinstance(metrics_dict, dict):
            metrics_dict = {"metrics": metrics_dict}
        metrics_dict = {prefix+k: {"format": format, "data": [v]} for k, v in metrics_dict.items()}
        return metrics_dict



from typing import Dict, Any, Union
from rainbowneko.data.label_loader import BaseLabelLoader, auto_label_loader
from rainbowneko.utils import is_image_file
from webdataset import DataPipeline
import webdataset as wds

from .base import DataSource


class WebDatasetSource(DataSource):
    def __init__(self, pipeline: DataPipeline, repeat=1, **kwargs):
        super().__init__(repeat, **kwargs)
        self.pipeline = pipeline

    def __getitem__(self, index) -> Dict[str, Any]:
        raise NotImplementedError('WebDatasetSource is not indexable')

    def __iter__(self):
        self.pipeline_iter = iter(self.pipeline)
        return self

    def __next__(self):
        data = next(self.pipeline_iter)
        return {
            'id': data['__key__'],
            **{k: v for k, v in data.items() if not k.startswith('__')}
        }


class WebDSImageLabelSource(WebDatasetSource):
    '''
    data.tar:
        - {id1}.{jpg|png|webp|...}
        - {id2}.{jpg|png|webp|...}

    label:
        id1: label1
        id2: label2
    '''

    def __init__(self, pipeline: DataPipeline, label_file, repeat=1, **kwargs):
        super().__init__(pipeline, repeat, **kwargs)
        self.label_dict = self._load_label_data(label_file)

    def _load_label_data(self, label_file: Union[str, BaseLabelLoader]):
        if label_file is None:
            return {}
        elif isinstance(label_file, str):
            return auto_label_loader(label_file).load()
        else:
            return label_file.load()

    def __next__(self):
        data = next(self.pipeline_iter)
        img_id = data['__key__']
        image = [v for k, v in data.items() if is_image_file(k)][0]

        return {
            'id': img_id,
            'image': image,
            'label': self.label_dict.get(img_id, None),
        }


def image_pipeline(url, buffer_size=1000):
    return wds.DataPipeline(
        wds.SimpleShardList(url),
        # at this point we have an iterator over all the shards

        # this shuffles the shards
        wds.shuffle(buffer_size),

        # add wds.split_by_node here if you are using multiple nodes
        wds.split_by_worker,

        # at this point, we have an iterator over the shards assigned to each worker
        wds.tarfile_to_samples(),
        wds.split_by_node,

        wds.decode("pil"),
    )



import os
from typing import Dict, Any

from PIL import Image
from loguru import logger
from tqdm.auto import tqdm

from .cli_logger import CLILogger

class TQDMLogger(CLILogger):
    def __init__(self, exp_dir, out_path, log_step=10,
                 img_log_dir=None, img_ext='png', img_quality=95):
        super().__init__(exp_dir, None, log_step, img_log_dir=img_log_dir, img_ext=img_ext, img_quality=img_quality)
        self.out_path = out_path
        self.pbar = tqdm()
        if exp_dir is not None:  # exp_dir is only available in local main process
            pass
        else:
            self.disable()

    def enable(self):
        super().enable()
        logger.enable("__main__")
        self.pbar.disable = False

    def disable(self):
        super().disable()
        logger.disable("__main__")
        self.pbar.disable = True

    def log_text(self, datas: Dict[str, Any], step: int = 0):
        step_key = None
        for k, v in datas.items():
            if os.path.basename(k).lower() == 'step':
                step_key = k
                if self.pbar.total is None:
                    self.pbar.total = v['data'][1]
                    break

        if step_key is not None:
            del datas[step_key]

        desc = ', '.join([f"{os.path.basename(k)} = {v['format'].format(*v['data'])}" for k, v in datas.items()])
        self.pbar.n = step
        self.pbar.last_print_n = step
        self.pbar.refresh()
        self.pbar.set_description(desc)




from typing import Dict, Any, List

from PIL import Image

class BaseLogger:
    def __init__(self, exp_dir, out_path, log_step=10):
        self.exp_dir = exp_dir
        self.out_path = out_path
        self.log_step = log_step
        self.enable_log = True

        self.evaluator_list = []

    def enable(self):
        self.enable_log = True

    def disable(self):
        self.enable_log = False

    def add_evaluator(self, evaluator):
        self.evaluator_list.append(evaluator)

    def info(self, info):
        if self.enable_log:
            self._info(info)

    def _info(self, info):
        raise NotImplementedError()

    def log(self, datas: Dict[str, Any], step: int = 0):
        text_dict = {}
        img_dict = {}
        for k, v in datas.items():
            if isinstance(v, Image.Image):
                img_dict[k] = v
            else:
                text_dict[k] = v

        if self.enable_log:
            if len(text_dict)>0:
                self.log_text(text_dict, step)
            if len(img_dict)>0:
                self.log_image(img_dict, step)

    def log_text(self, datas: Dict[str, Any], step: int = 0):
        raise NotImplementedError()

    def log_image(self, imgs: Dict[str, Image.Image], step: int = 0):
        raise NotImplementedError()

class LoggerGroup:
    def __init__(self, logger_list: List[BaseLogger]):
        self.logger_list = logger_list

    def enable(self):
        for logger in self.logger_list:
            logger.enable()

    def disable(self):
        for logger in self.logger_list:
            logger.disable()

    def add_evaluator(self, evaluator):
        for logger in self.logger_list:
            logger.add_evaluator(evaluator)

    def info(self, info):
        for logger in self.logger_list:
            logger.info(info)

    def log(self, datas: Dict[str, Any], step: int = 0):
        for logger in self.logger_list:
            if step % logger.log_step == 0:
                logger.log(datas, step)

    def __len__(self):
        return len(self.logger_list)



from typing import Union, List

from .base import NekoLoader, NekoSaver, LAYERS_ALL, LAYERS_TRAINABLE
from .ckpt import NekoModelSaver, NekoPluginSaver, NekoModelLoader, NekoPluginLoader, NekoEasySaver
from .format import PKLFormat, SafeTensorFormat, CkptFormat
from .model import NekoPluginModuleSaver, NekoModelModuleSaver
from .source import LocalCkptSource
from .resume import NekoResumer


def ckpt_saver(ckpt_type='safetensors', layers=LAYERS_ALL, state_prefix='', target_module: Union[str, List[str]] = '',
               prefix=None):
    if ckpt_type == 'pkl':
        format = PKLFormat()
    elif ckpt_type == 'safetensors':
        format = SafeTensorFormat()
    else:
        raise ValueError(f'Unknown ckpt_type: {ckpt_type}')

    return NekoModelSaver(format, LocalCkptSource(), layers=layers, state_prefix=state_prefix, target_module=target_module,
                          prefix=prefix)


def plugin_saver(ckpt_type='safetensors', layers=LAYERS_ALL, state_prefix='', target_plugin: Union[str, List[str]] = '',
                 prefix=None, plugin_from_raw=False):
    if ckpt_type == 'pkl':
        format = PKLFormat()
    elif ckpt_type == 'safetensors':
        format = SafeTensorFormat()
    else:
        raise ValueError(f'Unknown ckpt_type: {ckpt_type}')

    return NekoPluginSaver(format, LocalCkptSource(), layers=layers, state_prefix=state_prefix, target_plugin=target_plugin,
                          prefix=prefix, plugin_from_raw=plugin_from_raw)

def ckpt_easy_saver(ckpt_type='safetensors', layers=LAYERS_ALL, state_prefix='', prefix=None):
    if ckpt_type == 'pkl':
        format = PKLFormat()
    elif ckpt_type == 'safetensors':
        format = SafeTensorFormat()
    else:
        raise ValueError(f'Unknown ckpt_type: {ckpt_type}')

    return NekoEasySaver(format, LocalCkptSource(), layers=layers, state_prefix=state_prefix, prefix=prefix)


def auto_ckpt_loader(path: str, layers=LAYERS_ALL, target_module='',
                     state_prefix=None, base_model_alpha=0.0, alpha=1.0, load_ema=False):
    if path.endswith('.safetensors'):
        format = SafeTensorFormat()
    else:
        format = PKLFormat()

    return NekoModelLoader(format, LocalCkptSource(), path=path, layers=layers, target_module=target_module,
                           state_prefix=state_prefix, base_model_alpha=base_model_alpha, alpha=alpha, load_ema=load_ema)


def auto_plugin_loader(path: str, layers=LAYERS_ALL, target_plugin='', state_prefix=None, base_model_alpha=0.0, load_ema=False,
                       **plugin_kwargs):
    if path.endswith('.safetensors'):
        format = SafeTensorFormat()
    else:
        format = PKLFormat()

    return NekoPluginLoader(format, LocalCkptSource(), path=path, layers=layers, target_plugin=target_plugin,
                            state_prefix=state_prefix, base_model_alpha=base_model_alpha, load_ema=load_ema, **plugin_kwargs)


def auto_load_ckpt(path: str, **kwargs):
    loader = auto_ckpt_loader(path=path, **kwargs)
    return loader.load(path)



import math

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F


class BatchCosineSimilarity(nn.Module):
    def __init__(self):
        super().__init__()
        self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

    def forward(self, image_features):  # x: BxN
        # normalized features
        image_features = image_features / image_features.norm(dim=1, keepdim=True)

        # cosine similarity as logits
        logit_scale = self.logit_scale.exp()
        logits_per_image = logit_scale * torch.mm(image_features, image_features.transpose(0, 1))
        # if self.training:
        logits_per_image = logits_per_image - torch.diag_embed(torch.diag(logits_per_image))

        return logits_per_image


class StyleSimilarity(nn.Module):
    def __init__(self, batch_mean=False):
        super().__init__()
        self.batch_mean = batch_mean

    def forward(self, feat1, feat2):  # feat1: [B,C,H,W]
        sim = F.mse_loss(self.gram_matrix(feat1), self.gram_matrix(feat2), reduction='none')
        if self.batch_mean:
            return sim.mean(dim=(1, 2))*1000
        else:
            return sim.mean()*1000

    @staticmethod
    def gram_matrix(x, should_normalize=True):
        (b, c, h, w) = x.size()
        features = x.view(b, c, h * w)
        if should_normalize:
            features = features / math.sqrt(c * h * w)
        features_t = features.transpose(1, 2)
        gram = features.bmm(features_t)  # [B,C,C]
        return gram



from typing import Union, Dict, Any

from rainbowneko.data.label_loader import BaseLabelLoader, auto_label_loader
from rainbowneko.utils import is_image_file, Path_Like

from .base import VisionDataSource


class ImageLabelSource(VisionDataSource):
    def __init__(self, img_root: Path_Like, label_file, repeat=1, **kwargs):
        super().__init__(img_root, repeat=repeat)

        self.label_dict = self._load_label_data(label_file)
        self.img_ids = self._load_img_ids(self.label_dict)

    def _load_img_ids(self, label_dict):
        return [x for x in label_dict.keys() if is_image_file(x)] * self.repeat

    def _load_label_data(self, label_file: Union[str, BaseLabelLoader]):
        if label_file is None:
            return {}
        elif isinstance(label_file, str):
            return auto_label_loader(label_file).load()
        else:
            return label_file.load()

    def __len__(self):
        return len(self.img_ids)

    def __getitem__(self, index) -> Dict[str, Any]:
        img_id = self.img_ids[index]
        path = self.img_root / img_id
        return {
            'id': img_id,
            'image': path,
            'label': self.label_dict.get(self.img_ids[index], None)
        }



from .metrics import BaseMetric, MetricContainer, FullMetricContainer, MetricGroup
from .evaluator import Evaluator, EvaluatorGroup, WorkflowEvaluator
from .torchmetrics import TorchMetricsContainer
from .preview import Previewer, WorkflowPreviewer



from rainbowneko.parser import CfgModelParser, neko_cfg

@neko_cfg
def make_cfg():
    return dict(
        model_part=CfgModelParser([
            dict(
                lr=1e-6,
                layers=[''],  # train all layers
            )
        ]),

        model_plugin=None,
    )



import torch.nn as nn
import torch.nn.functional as F

class SoftCELoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        self.reduction = reduction

    def forward(self, logits, soft_targets):
        log_probabilities = F.log_softmax(logits, dim=-1)
        loss = F.kl_div(log_probabilities, soft_targets, reduction=self.reduction)
        return loss



from typing import Dict, Any, Union

import torch
from torch.serialization import FILE_LIKE


class CkptFormat:
    EXT = ''

    def save_ckpt(self, sd_model: Union[torch.nn.Module, Dict[str, Any]], save_f: FILE_LIKE):
        raise NotImplementedError

    def load_ckpt(self, ckpt_f: FILE_LIKE, map_location="cpu"):
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}(EXT={self.EXT})'



import torch
import torchvision
import torchvision.transforms as T
from torch import nn

from rainbowneko.data import IndexSource, HandlerChain, LoadImageHandler, ImageHandler, BaseDataset, BaseBucket
from rainbowneko.infer import HandlerAction, DataLoaderAction
from rainbowneko.infer.workflow import (Actions, BuildModelAction, PrepareAction, FeedAction, ForwardAction,
                                        LambdaAction, VisClassAction, LoadModelAction)
from rainbowneko.models.wrapper import SingleWrapper
from rainbowneko.parser import neko_cfg
from rainbowneko.ckpt_manager import auto_ckpt_loader

num_classes = 10

def load_resnet():
    model = torchvision.models.resnet18(weights=torchvision.models.ResNet18_Weights.DEFAULT)
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    return model

@neko_cfg
def infer_one(path):
    return Actions([
        FeedAction(image=path),
        HandlerAction(handler=HandlerChain(
            load=LoadImageHandler(),
            image=ImageHandler(transform=T.Compose([
                T.CenterCrop(size=32),
                T.ToTensor(),
                T.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010]),
            ]))
        ), key_map_in=('image -> image',)),
        LambdaAction(f_act=lambda image, **kwargs: {'image': image.unsqueeze(0)}),
        ForwardAction(key_map_in=('image -> input.image', 'model -> model')),
        VisClassAction(
            class_map=['airplanes', 'cars', 'birds', 'cats', 'deer', 'dogs', 'frogs', 'horses', 'ships', 'trucks'],
            key_map_in=('output.pred -> pred',)
        )
    ])

@neko_cfg
def infer_all(path):
    return DataLoaderAction(
        dataset=BaseDataset(_partial_=True, batch_size=1, loss_weight=1.0,
            source=dict(
                data_source1=IndexSource(
                    data=torchvision.datasets.cifar.CIFAR10(root=path, train=False, download=True)
                ),
            ),
            handler=HandlerChain(
                load=LoadImageHandler(),
                image=ImageHandler(transform=T.Compose([
                        T.CenterCrop(size=32),
                        T.ToTensor(),
                        T.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010]),
                    ]),
                )
            ),
            bucket=BaseBucket(),
        ),
        actions=Actions([
            ForwardAction(key_map_in=('image -> input.image', 'model -> model', 'device -> device', 'dtype -> dtype')),
            VisClassAction(
                class_map=['airplanes', 'cars', 'birds', 'cats', 'deer', 'dogs', 'frogs', 'horses', 'ships', 'trucks'],
                key_map_in=('output.pred -> pred',)
            )
        ])
    )

@neko_cfg
def make_cfg():
    return dict(workflow=Actions(actions=[
        PrepareAction(device='cpu', dtype=torch.float16),
        BuildModelAction(SingleWrapper(_partial_=True, model=load_resnet())),
        LoadModelAction(dict(
            model=auto_ckpt_loader(
                target_module='model',
                path='exps/cifar/ckpts/cifar-resnet18-3900.ckpt',
            ),
        )),
        #infer_one(path=r"E:\dataset\frog10.png")
        infer_all(path=r'D:\others\dataset\cifar')
    ]))


from typing import Dict, Union, List

from torch import nn

from rainbowneko.models.plugin import PluginGroup, BasePluginBlock
from .base import NekoLoader, NekoSaver, LAYERS_ALL, LAYERS_TRAINABLE
from .format import CkptFormat
from .locator import get_match_layers
from .source import LocalCkptSource


class NekoModelLoader(NekoLoader):
    def __init__(self, format: CkptFormat = None, source: LocalCkptSource = None, path: str = None, layers='all',
                 target_module='',
                 state_prefix=None, base_model_alpha=0.0, alpha=1.0, load_ema=False):
        super().__init__(format=format, source=source, layers=layers)
        self.path = path

        self.target_module = target_module
        self.base_model_alpha = base_model_alpha
        self.alpha = alpha
        self.load_ema = load_ema
        self.state_prefix = state_prefix

    def load_to(self, name, model):
        model = model if self.target_module == '' else eval(f"model.{self.target_module}")
        named_params = {k: v for k, v in model.named_parameters()}
        named_params.update({k: v for k, v in model.named_buffers()})

        part_state = self.load(self.path, map_location='cpu')
        if self.load_ema:
            part_state = part_state['base_ema']
        else:
            if 'base' in part_state:
                part_state = part_state['base']

        if self.state_prefix:
            state_prefix_len = len(self.state_prefix)
            part_state = {k[state_prefix_len:]: v for k, v in part_state.items() if k.startswith(self.state_prefix)}

        if self.layers == LAYERS_ALL:
            for k, v in part_state.items():
                named_params[k].data = self.base_model_alpha * named_params[k].data + self.alpha * v.to(
                    named_params[k].data.device)
        else:
            named_modules = {k: v for k, v in model.named_modules()}
            match_blocks = get_match_layers(self.layers, named_modules)
            state_add = {k: v for blk in match_blocks for k, v in part_state.items() if k.startswith(blk)}
            for k, v in state_add.items():
                named_params[k].data = self.base_model_alpha * named_params[k].data + self.alpha * v.to(
                    named_params[k].data.device)


class NekoModelSaver(NekoSaver):
    def __init__(self, format: CkptFormat = None, source: LocalCkptSource = None, layers='all', state_prefix='',
                 target_module: Union[str, List[str]] = '', prefix=None):
        super().__init__(format=format, source=source, layers=layers, state_prefix=state_prefix)
        self.prefix = prefix
        if isinstance(target_module, str):
            target_module = [target_module]
        self.target_module = target_module

    def exclude_state(self, state, key):
        if key is None:
            return state
        else:
            return {k: v for k, v in state.items() if key not in k}

    def save_to(self, name, model: nn.Module, plugin_groups: Dict[str, PluginGroup], model_ema=None, exclude_key=None,
                name_template=None):
        sd_base = {}
        for item in self.target_module:
            block = model if item == '' else eval(f"model.{item}")
            sd_item = self.exclude_state(
                BasePluginBlock.extract_state_without_plugin(block, trainable=self.layers == LAYERS_TRAINABLE), exclude_key
            )

            # filter layers
            if not isinstance(self.layers, str):
                named_modules = {k: v for k, v in model.named_modules()}
                match_blocks = get_match_layers(self.layers, named_modules)
                sd_item = {k: v for blk in match_blocks for k, v in sd_item.items() if k.startswith(blk)}

            sd_base.update(sd_item)

        if len(sd_base) > 0:
            sd_model = {"base": sd_base}
            if model_ema is not None:
                sd_ema = model_ema.state_dict()
                sd_ema = {k: sd_ema[k] for k in sd_base.keys()}
                sd_model["base_ema"] = self.exclude_state(sd_ema, exclude_key)
                sd_model["base_ema"] = self.clean_prefix(sd_model["base_ema"])
            sd_model["base"] = self.clean_prefix(sd_model["base"])
            if name_template is not None:
                name = name_template.format(name)
            self.save(sd_model, name, prefix=self.prefix)


class NekoPluginLoader(NekoLoader):
    def __init__(self, format: CkptFormat = None, source: LocalCkptSource = None, path: str = None, layers='all',
                 target_plugin=None,
                 state_prefix=None, base_model_alpha=0.0, load_ema=False, **plugin_kwargs):
        super().__init__(format=format, source=source, layers=layers)
        self.path = path

        self.target_plugin = target_plugin
        self.state_prefix = state_prefix
        self.base_model_alpha = base_model_alpha
        self.load_ema = load_ema

        self.plugin_kwargs = plugin_kwargs

    def load_to(self, name: str, plugin_groups: Dict[str, PluginGroup]):
        # get plugin_group to load
        plugin_group = plugin_groups[self.target_plugin or name]
        plugin_state = self.load(self.path, map_location='cpu')['base_ema' if self.load_ema else 'base']

        if self.state_prefix:
            state_prefix_len = len(self.state_prefix)
            plugin_state = {k[state_prefix_len:]: v for k, v in plugin_state.items() if k.startswith(self.state_prefix)}

        # filter layers to load
        plugin_dict = plugin_group.plugin_dict
        if self.layers != LAYERS_ALL:
            plugin_dict = get_match_layers(self.layers, plugin_dict)

        # state to plugin_block_state
        plugin_block_state = {}
        for pname, p in plugin_state.items():
            prefix, block_name = pname.split('.___.', 1)
            plugin_block_state.setdefault(prefix, {})[block_name] = p

        # Load state to plugin
        missing_keys = []
        unexpected_keys = []
        for key, plugin in plugin_dict.items():
            if key not in plugin_block_state:
                missing_keys.append(key)
            else:
                load_info = plugin.load_state_dict(plugin_block_state.pop(key), strict=False)
                plugin.set_hyper_params(**self.plugin_kwargs)
                missing_keys.extend(load_info.missing_keys)
                unexpected_keys.extend(load_info.unexpected_keys)


class NekoPluginSaver(NekoSaver):
    def __init__(self, format: CkptFormat = None, source: LocalCkptSource = None, layers='all', state_prefix='',
                 target_plugin: Union[str, List[str]] = '', prefix=None, plugin_from_raw=False):
        super().__init__(format=format, source=source, layers=layers, state_prefix=state_prefix)
        self.prefix = prefix
        if isinstance(target_plugin, str):
            target_plugin = [target_plugin]
        self.target_plugin = target_plugin
        self.plugin_from_raw = plugin_from_raw

    def save_to(self, name, host_model, plugin_groups: Dict[str, PluginGroup], model_ema=None, exclude_key=None,
                name_template=None):
        sd_base = {}
        sd_ema = {}
        for plugin_name in self.target_plugin:
            plugin = plugin_groups[plugin_name]
            sd_item = plugin.state_dict(host_model if self.plugin_from_raw else None)

            # filter layers
            if not isinstance(self.layers, str):
                match_blocks = get_match_layers(self.layers, plugin.plugin_dict)
                sd_item = {k: v for blk in match_blocks for k, v in sd_item.items() if k.startswith(blk)}

            sd_base.update(sd_item)

            if model_ema is not None:
                sd_ema_item = plugin.state_dict(model_ema)
                sd_ema_item = {k: sd_ema_item[k] for k in sd_item.keys()}
                sd_ema.update(sd_ema_item)

        if len(sd_base) > 0:
            sd_plugin = {'base': sd_base}
            if model_ema is not None:
                sd_plugin["base_ema"] = self.clean_prefix(sd_ema)
            sd_plugin["base"] = self.clean_prefix(sd_plugin["base"])
            if name_template is not None:
                name = name_template.format(name)
            self.save(sd_plugin, name, prefix=self.prefix)


class NekoEasySaver(NekoSaver):
    def __init__(self, format: CkptFormat = None, source: LocalCkptSource = None, layers='all', state_prefix='',
                 prefix=None, plugin_from_raw=False):
        super().__init__(format=format, source=source, layers=layers, state_prefix=state_prefix)

        self.prefix = prefix
        self.plugin_from_raw = plugin_from_raw

    def exclude_state(self, state, key):
        if key is None:
            return state
        else:
            return {k: v for k, v in state.items() if key not in k}

    def model_save_to(self, name, model: nn.Module, model_ema=None, exclude_key=None,
                      name_template=None):
        sd_base = {}
        sd_item = self.exclude_state(
            BasePluginBlock.extract_state_without_plugin(model, trainable=self.layers == LAYERS_TRAINABLE), exclude_key
        )

        # filter layers
        if not isinstance(self.layers, str):
            named_modules = {k: v for k, v in model.named_modules()}
            match_blocks = get_match_layers(self.layers, named_modules)
            sd_item = {k: v for blk in match_blocks for k, v in sd_item.items() if k.startswith(blk)}

        sd_base.update(sd_item)

        if len(sd_base) > 0:
            sd_model = {"base": sd_base}
            if model_ema is not None:
                sd_ema = model_ema.state_dict()
                sd_ema = {k: sd_ema[k] for k in sd_base.keys()}
                sd_model["base_ema"] = self.exclude_state(sd_ema, exclude_key)
                sd_model["base_ema"] = self.clean_prefix(sd_model["base_ema"])
            sd_model["base"] = self.clean_prefix(sd_model["base"])
            if name_template is not None:
                name = name_template.format(name)
            self.save(sd_model, name, prefix=self.prefix)

    def plugin_save_to(self, name, host_model, plugin_groups: Dict[str, PluginGroup], model_ema=None, exclude_key=None,
                       name_template=None):
        for plugin_name, plugin in plugin_groups:
            sd_base = plugin.state_dict(host_model if self.plugin_from_raw else None)

            # filter layers
            # TODO: filter trainable
            if not isinstance(self.layers, str):
                match_blocks = get_match_layers(self.layers, plugin.plugin_dict)
                sd_base = {k: v for blk in match_blocks for k, v in sd_base.items() if k.startswith(blk)}

            if model_ema is not None:
                sd_ema = plugin.state_dict(model_ema)
                sd_ema = {k: sd_ema[k] for k in sd_base.keys()}

            if len(sd_base) > 0:
                sd_plugin = {'base': sd_base}
                if model_ema is not None:
                    sd_plugin["base_ema"] = self.clean_prefix(sd_ema)
                sd_plugin["base"] = self.clean_prefix(sd_plugin["base"])
                if name_template is not None:
                    name = name_template.format(f'{name}-{plugin_name}')
                self.save(sd_plugin, name, prefix=self.prefix)

    def save_to(self, name, host_model, plugin_groups: Dict[str, PluginGroup], model_ema=None, exclude_key=None,
                name_template=None):
        self.model_save_to(name, host_model, model_ema, exclude_key, name_template)
        self.plugin_save_to(name, host_model, plugin_groups, model_ema, exclude_key, name_template)



import unittest
import torch

torch_device = "cuda" if torch.cuda.is_available() else "cpu"

def require_torch_gpu(test_case):
    """Decorator marking a test that requires CUDA and PyTorch."""
    return unittest.skipUnless(torch_device == "cuda", "test requires PyTorch+CUDA")(
        test_case
    )


# These decorators are for accelerator-specific behaviours that are not GPU-specific
def require_torch_accelerator(test_case):
    """Decorator marking a test that requires an accelerator backend and PyTorch."""
    return unittest.skipUnless(torch_device != "cpu", "test requires accelerator+PyTorch")(
        test_case
    )


def require_torch_multi_gpu(test_case):
    """
    Decorator marking a test that requires a multi-GPU setup (in PyTorch). These tests are skipped on a machine without
    multiple GPUs. To run *only* the multi_gpu tests, assuming all test names contain multi_gpu: $ pytest -sv ./tests
    -k "multi_gpu"
    """
    return unittest.skipUnless(torch.cuda.device_count() > 1, "test requires multiple GPUs")(test_case)


import torch
from torch.utils.data.distributed import DistributedSampler
from typing import Iterator
import platform
import math

class DistributedCycleSampler(DistributedSampler):
    _cycle = 1000

    def __iter__(self) -> Iterator:
        def _iter():
            while True:
                if self.shuffle:
                    # deterministically shuffle based on epoch and seed
                    g = torch.Generator()
                    g.manual_seed(self.seed + self.epoch)
                    indices = torch.randperm(len(self.dataset), generator=g).tolist()  # type: ignore[arg-type]
                else:
                    indices = list(range(len(self.dataset)))  # type: ignore[arg-type]

                if not self.drop_last:
                    # add extra samples to make it evenly divisible
                    padding_size = self.total_size - len(indices)
                    if padding_size <= len(indices):
                        indices += indices[:padding_size]
                    else:
                        indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
                else:
                    # remove tail of data to make it evenly divisible.
                    indices = indices[:self.total_size]
                assert len(indices) == self.total_size

                # subsample
                indices = indices[self.rank:self.total_size:self.num_replicas]
                assert len(indices) == self.num_samples

                for idx in indices:
                    yield idx
                self.epoch+=1

                if self.epoch>=self._cycle:
                    break

        return _iter()

    def __len__(self) -> int:
        return self.num_samples #*self._cycle

def get_sampler(train=True):
    # Fix DataLoader frequently reload bugs in windows
    if train and platform.system().lower() == 'windows':
        return DistributedCycleSampler
    else:
        return DistributedSampler


from .base_logger import BaseLogger, LoggerGroup
from .cli_logger import CLILogger
from .tqdm_logger import TQDMLogger

try:
    from .tensorboard_logger import TBLogger
except:
    print('INFO: tensorboard is not available')

try:
    from .wandb_logger import WanDBLogger
except:
    print('INFO: wandb is not available')


from concurrent.futures import ThreadPoolExecutor

import numpy as np
from loguru import logger
from rainbowneko.utils.img_size_tool import get_image_size
from sklearn.cluster import KMeans
from tqdm.auto import tqdm

from .ratio import RatioBucket


class LongEdgeBucket(RatioBucket):
    def __init__(self, target_edge=640, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None):
        super().__init__(step_size=step_size, num_bucket=num_bucket, pre_build_bucket=pre_build_bucket)
        self.target_edge = target_edge

    def build_buckets_from_images(self):
        '''
        根据图像尺寸聚类，不会resize图像，只有剪裁和填充操作。
        '''
        logger.info('build buckets from images size')

        def get_size(info):
            data, source = info
            w, h = get_image_size(data)
            scale = self.target_edge / max(w, h)
            return round(w * scale), round(h * scale)

        size_list = []
        with self.source.return_source(), ThreadPoolExecutor() as executor:
            for w, h in tqdm(executor.map(get_size, self.source), desc='get image info', total=len(self.source)):
                size_list.append([w, h])
        size_list = np.array(size_list)

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514, verbose=True).fit(size_list)
        labels = kmeans.labels_
        size_buckets = kmeans.cluster_centers_

        # SD需要边长是8的倍数
        self.size_buckets = (np.round(size_buckets / self.step_size) * self.step_size).astype(int)

        if self.source_indexable:
            self.buckets = []  # [bucket_id:[file_idx,...]]
            self.idx_bucket_map = labels
            for bidx in range(self.num_bucket):
                self.buckets.append(np.where(labels == bidx)[0].tolist())
            logger.info('buckets info: ' + ', '.join(
                f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))
        else:
            logger.info('buckets info: ' + ', '.join(f'size:{size}' for size in enumerate(self.size_buckets)))

    def next_data(self, shuffle=True):
        def assign_bucket(datas, source, buckets):
            w, h = source.get_image_size(datas)
            scale = self.target_edge / max(w, h)
            w, h = round(w * scale), round(h * scale)
            bucket_idx = np.linalg.norm(self.size_buckets-np.array([[w,h]]), axis=1).argmin()
            datas['image_size'] = self.size_buckets[bucket_idx]
            buckets[bucket_idx].append(datas)

        if not hasattr(self, 'buffer_iter'):
            self.buffer_iter = self._buffer(self.bs, assign_bucket, rs=self.rs if shuffle else None)
        return next(self.buffer_iter)

    @classmethod
    def from_files(cls, target_edge, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None, **kwargs):
        arb = cls(target_edge, step_size, num_bucket, pre_build_bucket=pre_build_bucket)
        arb._build = arb.build_buckets_from_images
        return arb



from functools import partial

import torch
import torchvision
import torchvision.transforms as T
from torch import nn
from torch.nn import CrossEntropyLoss
from torchmetrics.classification import MulticlassAccuracy, MulticlassF1Score

from cfgs.py.train import train_base, tuning_base
from rainbowneko.ckpt_manager import ckpt_saver
from rainbowneko.data import BaseDataset
from rainbowneko.data import FixedBucket
from rainbowneko.data.handler import HandlerChain, ImageHandler, LoadImageHandler
from rainbowneko.data.source import IndexSource
from rainbowneko.evaluate import MetricGroup, MetricContainer, Evaluator
from rainbowneko.models.wrapper import SingleWrapper
from rainbowneko.parser import CfgWDModelParser, neko_cfg
from rainbowneko.train.loss import LossContainer
from rainbowneko.utils import CosineLR

num_classes = 10

def load_resnet():
    model = torchvision.models.resnet18(weights=torchvision.models.ResNet18_Weights.DEFAULT)
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    return model

@neko_cfg
def make_cfg():
    return dict(
        _base_=[train_base, tuning_base],

        model_part=CfgWDModelParser([
            dict(
                lr=1e-4,
                layers=[''],  # train all layers
            )
        ]),

        # func(_partial_=True, ...) same as partial(func, ...)
        ckpt_saver=dict(
            model=ckpt_saver(target_module='model')
        ),

        train=dict(
            train_epochs=10,
            workers=2,
            max_grad_norm=None,
            save_step=2000,

            loss=LossContainer(loss=CrossEntropyLoss()),

            optimizer=partial(torch.optim.AdamW, weight_decay=5e-4),

            scale_lr=False,
            scheduler=CosineLR(
                _partial_=True,
                warmup_steps=100,
            ),
            metrics=MetricGroup(
                acc=MetricContainer(MulticlassAccuracy(num_classes=num_classes)),
                f1=MetricContainer(MulticlassF1Score(num_classes=num_classes)),
            ),
        ),

        model=dict(
            name='cifar-resnet18',
            wrapper=partial(SingleWrapper, model=load_resnet())
        ),

        data_train=cfg_data(), # config can be split into another function with @neko_cfg

        evaluator=cfg_evaluator(),
    )

@neko_cfg
def cfg_data():
    return dict(
        dataset1=partial(BaseDataset, batch_size=16, loss_weight=1.0,
            source=dict(
                data_source1=IndexSource(
                    data=torchvision.datasets.cifar.CIFAR10(root=r'D:\others\dataset\cifar', train=True, download=True)
                ),
            ),
            handler=HandlerChain(
                load=LoadImageHandler(),
                bucket=FixedBucket.handler, # bucket 会自带一些处理模块
                image=ImageHandler(transform=T.Compose([
                        T.RandomCrop(size=32, padding=4),
                        T.RandomHorizontalFlip(),
                        T.ToTensor(),
                        T.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010]),
                    ]),
                )
            ),
            bucket=FixedBucket(target_size=32),
        )
    )

@neko_cfg
def cfg_evaluator():
    return partial(Evaluator,
        interval=50,
        metric=MetricGroup(
            acc=MetricContainer(MulticlassAccuracy(num_classes=num_classes)),
            f1=MetricContainer(MulticlassF1Score(num_classes=num_classes)),
        ),
        dataset=partial(BaseDataset, batch_size=16, loss_weight=1.0,
            source=dict(
                data_source1=IndexSource(
                    data=torchvision.datasets.cifar.CIFAR10(root=r'D:\others\dataset\cifar', train=False, download=True)
                ),
            ),
            handler=HandlerChain(
                load=LoadImageHandler(),
                bucket=FixedBucket.handler,
                image=ImageHandler(transform=T.Compose([
                        T.ToTensor(),
                        T.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010]),
                    ]),
                )
            ),
            bucket=FixedBucket(target_size=32),
        )
    )


import functools
from typing import Dict, List, Tuple

from rainbowneko.models.plugin import SinglePluginBlock, MultiPluginBlock, PluginBlock, PluginGroup, PatchPluginBlock
from rainbowneko.utils import net_path_join, split_module_name
from torch import nn

from rainbowneko.ckpt_manager.locator import get_match_layers


class CfgPluginParser:
    def __init__(self, cfg_plugin: Dict, lr=1e-5, weight_decay=0):
        self.cfg_plugin = cfg_plugin
        self.lr = lr
        self.weight_decay = weight_decay

    def get_params_multi(self, builder, named_modules, plugin_name, model, train_plugin):
        blocks = {}
        params_list = []

        from_layers = [
            {**item, "layer": named_modules[item["layer"]]}
            for item in get_match_layers(builder.keywords.pop("from_layers"), named_modules, return_metas=True)
        ]
        to_layers = [
            {**item, "layer": named_modules[item["layer"]]}
            for item in get_match_layers(builder.keywords.pop("to_layers"), named_modules, return_metas=True)
        ]

        layer = builder(name=plugin_name, host_model=model, from_layers=from_layers, to_layers=to_layers)
        if train_plugin:
            layer.train()
            params = layer.get_trainable_parameters()
            for p in params:
                p.requires_grad_(True)
                params_list.append(p)
        else:
            layer.requires_grad_(False)
            layer.eval()
        blocks[""] = layer
        return blocks, params_list

    def get_params_single(self, builder, named_modules, plugin_name, model, train_plugin):
        blocks = {}
        params_list = []

        layers_name = builder.keywords.pop("layers")
        for layer_name in get_match_layers(layers_name, named_modules):
            blocks = builder(name=plugin_name, host_model=model, host=named_modules[layer_name])
            if not isinstance(blocks, dict):
                blocks = {"": blocks}

            for k, v in blocks.items():
                blocks[net_path_join(layer_name, k)] = v
                if train_plugin:
                    v.train()
                    params = v.get_trainable_parameters()
                    for p in params:
                        p.requires_grad_(True)
                        params_list.append(p)
                else:
                    v.requires_grad_(False)
                    v.eval()

        return blocks, params_list

    def get_params_plugin(self, builder, named_modules, plugin_name, model, train_plugin):
        blocks = {}
        params_list = []

        from_layer = get_match_layers(builder.keywords.pop("from_layer"), named_modules, return_metas=True)
        to_layer = get_match_layers(builder.keywords.pop("to_layer"), named_modules, return_metas=True)

        for from_layer_meta, to_layer_meta in zip(from_layer, to_layer):
            from_layer_name = from_layer_meta["layer"]
            from_layer_meta["layer"] = named_modules[from_layer_name]
            to_layer_meta["layer"] = named_modules[to_layer_meta["layer"]]
            layer = builder(name=plugin_name, host_model=model, from_layer=from_layer_meta, to_layer=to_layer_meta)
            if train_plugin:
                layer.train()
                params = layer.get_trainable_parameters()
                for p in params:
                    p.requires_grad_(True)
                    params_list.append(p)
            else:
                layer.requires_grad_(False)
                layer.eval()
            blocks[from_layer_name] = layer

        return blocks, params_list

    def get_params_patch(self, builder, named_modules, plugin_name, model, train_plugin):
        blocks = {}
        params_list = []

        layers_name = builder.keywords.pop("layers")
        for layer_name in get_match_layers(layers_name, named_modules):
            parent_name, host_name = split_module_name(layer_name)
            layers = builder(
                name=plugin_name,
                host_model=model,
                host=named_modules[layer_name],
                parent_block=named_modules[parent_name],
                host_name=host_name,
            )
            if not isinstance(layers, dict):
                layers = {"": layers}

            for k, v in layers.items():
                blocks[net_path_join(layer_name, k)] = v
                if train_plugin:
                    v.train()
                    params = v.get_trainable_parameters()
                    for p in params:
                        p.requires_grad_(True)
                        params_list.append(p)
                else:
                    v.requires_grad_(False)
                    v.eval()

        return blocks, params_list

    def get_params_group(self, model) -> Tuple[List, Dict[str, PluginGroup]]:
        train_params = []
        all_plugin_group = {}

        if self.cfg_plugin is None:
            return train_params, all_plugin_group

        named_modules = {k: v for k, v in model.named_modules()}

        for plugin_name, builder in self.cfg_plugin.items():
            builder: functools.partial

            lr = builder.keywords.pop("lr") if "lr" in builder.keywords else self.lr
            train_plugin = builder.keywords.pop("train") if "train" in builder.keywords else True
            plugin_class = getattr(builder.func, "__self__", builder.func)  # support static or class method

            if issubclass(plugin_class, MultiPluginBlock):
                blocks, params_list = self.get_params_multi(builder, named_modules, plugin_name, model, train_plugin)
            elif issubclass(plugin_class, SinglePluginBlock):
                blocks, params_list = self.get_params_single(builder, named_modules, plugin_name, model, train_plugin)
            elif issubclass(plugin_class, PluginBlock):
                blocks, params_list = self.get_params_plugin(builder, named_modules, plugin_name, model, train_plugin)
            elif issubclass(plugin_class, PatchPluginBlock):
                blocks, params_list = self.get_params_patch(builder, named_modules, plugin_name, model, train_plugin)
            else:
                raise NotImplementedError(f"Unknown plugin: {plugin_class}")

            if train_plugin:
                train_params.append({"params": params_list, "lr": lr})
            all_plugin_group[plugin_name] = PluginGroup(blocks)
        return train_params, all_plugin_group

class CfgWDPluginParser(CfgPluginParser):
    def get_params_group(self, model) -> Tuple[List, Dict[str, PluginGroup]]:
        train_params = []
        all_plugin_group = {}

        if self.cfg_plugin is None:
            return train_params, all_plugin_group

        named_modules = {k: v for k, v in model.named_modules()}

        for plugin_name, builder in self.cfg_plugin.items():
            builder: functools.partial

            lr = builder.keywords.pop("lr") if "lr" in builder.keywords else self.lr
            weight_decay = builder.keywords.pop("weight_decay") if "weight_decay" in builder.keywords else self.weight_decay
            train_plugin = builder.keywords.pop("train") if "train" in builder.keywords else True
            plugin_class = getattr(builder.func, "__self__", builder.func)  # support static or class method

            if issubclass(plugin_class, MultiPluginBlock):
                blocks, params_list = self.get_params_multi(builder, named_modules, plugin_name, model, train_plugin)
            elif issubclass(plugin_class, SinglePluginBlock):
                blocks, params_list = self.get_params_single(builder, named_modules, plugin_name, model, train_plugin)
            elif issubclass(plugin_class, PluginBlock):
                blocks, params_list = self.get_params_plugin(builder, named_modules, plugin_name, model, train_plugin)
            elif issubclass(plugin_class, PatchPluginBlock):
                blocks, params_list = self.get_params_patch(builder, named_modules, plugin_name, model, train_plugin)
            else:
                raise NotImplementedError(f"Unknown plugin: {plugin_class}")

            if train_plugin:
                params_nowd, params_wd = self.get_wd(blocks, params_list)
                if len(params_nowd)>0:
                    train_params.append({"params": params_nowd, "lr": lr, "weight_decay": 0})
                if len(params_wd) > 0:
                    train_params.append({"params": params_wd, "lr": lr, "weight_decay": weight_decay})
            all_plugin_group[plugin_name] = PluginGroup(blocks)
        return train_params, all_plugin_group

    def get_wd(self, blocks: Dict[str, nn.Module], params_list: List[nn.Parameter]):
        params_nowd = []
        params_wd = []

        wd_params = set()
        for block in blocks.values():
            for m in block.modules():
                if isinstance(m, (nn.Linear, nn.Conv2d)):
                    wd_params.add(m.weight)

        for p in params_list:
            if p in wd_params:
                params_wd.append(p)
            else:
                params_nowd.append(p)
        return params_nowd, params_wd


import os
from typing import Dict, Any

import numpy as np
from PIL import Image
from torch.utils.tensorboard import SummaryWriter

from .base_logger import BaseLogger


class TBLogger(BaseLogger):
    def __init__(self, exp_dir, out_path, log_step=10):
        super().__init__(exp_dir, out_path, log_step)
        if exp_dir is not None:  # exp_dir is only available in local main process
            self.writer = SummaryWriter(os.path.join(exp_dir, out_path))
        else:
            self.writer = None
            self.disable()

    def _info(self, info):
        pass

    def log_text(self, datas: Dict[str, Any], step: int = 0):
        for k, v in datas.items():
            if len(v['data']) == 1:
                self.writer.add_scalar(k, v['data'][0], global_step=step)

    def log_image(self, imgs: Dict[str, Image.Image], step: int = 0):
        for name, img in imgs.items():
            self.writer.add_image(f'img/{name}', np.array(img), dataformats='HWC', global_step=step)



from .base import BasicAction
import torch
from typing import List

class VisClassAction(BasicAction):
    def __init__(self, class_map: List[str], key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.class_map = class_map

    def forward(self, pred: torch.Tensor, **states):
        for cls in pred.argmax(dim=-1):
            print(self.class_map[cls.item()])


from types import ModuleType
from typing import Dict, Union

import torch
import torch.distributed as dist
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from rainbowneko.infer import WorkflowRunner
from rainbowneko.models.wrapper import BaseWrapper
from rainbowneko.parser import load_config
from .metrics import BaseMetric, MetricGroup


class Evaluator:
    def __init__(self, trainer: "Trainer", data_loader: DataLoader, metric: BaseMetric, ds_name=None, interval=100, dataset:None=None):
        assert dataset is None, '"dataset" is a placeholder for cfg, should always be None.'
        self.data_loader = data_loader
        self.trainer = trainer
        self.metric = metric
        self.interval = interval
        self.ds_name = ds_name

        self.data_loader.dataset.bucket.rest(0)

        self.metric.to(self.device)

    @property
    def device(self):
        return self.trainer.device

    @property
    def dtype(self):
        return self.trainer.weight_dtype

    def forward_one_step(self, model, data):
        input_datas = {k: self.trainer.to_dev(v) for k, v in data.items() if k != "plugin_input"}
        if "plugin_input" in data:
            input_datas["plugin_input"] = {k: self.trainer.to_dev(v) for k, v in data["plugin_input"].items()}

        model_pred = model(self.ds_name, **input_datas)

        return model_pred, input_datas
    
    def cpu_gather(self, tensor):
        if self.trainer.world_size>1:
            if not hasattr(self, 'gloo_group'): # Transfer data on cpu
                self.gloo_group = dist.new_group(backend='gloo')

            world_size = dist.get_world_size()
            gathered_tensors = [torch.empty_like(tensor) for _ in range(world_size)]
            dist.all_gather(gathered_tensors, tensor, group=self.gloo_group)
            return torch.cat(gathered_tensors, dim=0)
        else:
            return tensor

    @torch.no_grad()
    def evaluate(self, step: int, model: BaseWrapper, prefix='eval/'):
        if step % self.interval != 0:
            return

        # record training layers
        training_layers = [layer for layer in model.modules() if layer.training]

        # reset metric
        model.eval()
        self.metric.reset()

        for data in tqdm(self.data_loader, disable=not self.trainer.is_local_main_process):
            pred, input_datas = self.forward_one_step(model, data)
            # update data to metric
            self.metric.update(pred, input_datas)

        v_metric = self.metric.finish(self.cpu_gather, self.trainer.is_local_main_process)
        if not isinstance(v_metric, dict):
            v_metric = {'metric': v_metric}

        data_size = len(self.data_loader.dataset)
        self.trainer.loggers.info(f'Evaluate: data size {data_size}')
        log_data = {
            "eval/Step": {
                "format": "{}",
                "data": [step],
            }
        }
        log_data.update(MetricGroup.format(v_metric, prefix=prefix))
        self.trainer.loggers.log(log_data, step)

        for layer in training_layers:
            layer.train()

    def to(self, device):
        self.metric.to(device)


class EvaluatorGroup:
    def __init__(self, loggers, evaluator_dict: Dict[str, Evaluator]):
        self.loggers = loggers
        self.evaluator_dict = evaluator_dict

    def evaluate(self, step: int, model: BaseWrapper):
        for name, evaluator in self.evaluator_dict.items():
            self.loggers.info(f'Evaluator {name}:')
            evaluator.evaluate(step, model, prefix=f'eval/{name}/')

    def to(self, device):
        for evaluator in self.evaluator_dict.values():
            evaluator.to(device)


class WorkflowEvaluator(Evaluator):
    def __init__(self, trainer: "Trainer", workflow: Union[str, ModuleType, Dict], ds_name=None, interval=100):
        self.trainer = trainer
        self.interval = interval
        self.ds_name = ds_name

        if isinstance(workflow, (ModuleType,str)):
            parser, conf = load_config(workflow)
            self.workflow_runner = WorkflowRunner(parser, conf)
        else:
            self.workflow_runner = WorkflowRunner(trainer.parser, workflow)

    @torch.no_grad()
    def evaluate(self, step: int, model: BaseWrapper, prefix='eval/'):
        if step % self.interval != 0:
            return

        # record training layers
        training_layers = [layer for layer in model.modules() if layer.training]

        model.eval()

        states = self.workflow_runner.run(model=model, in_preview=True, device=self.device, dtype=self.dtype,
                                          world_size=self.trainer.world_size, local_rank=self.trainer.local_rank)
        metric = states['_metric']

        v_metric = metric.finish(self.trainer.accelerator.gather, self.trainer.is_local_main_process)
        if not isinstance(v_metric, dict):
            v_metric = {'metric': v_metric}

        log_data = {
            "eval/Step": {
                "format": "{}",
                "data": [step],
            }
        }
        log_data.update(MetricGroup.format(v_metric, prefix=prefix))
        self.trainer.loggers.log(log_data, step)

        for layer in training_layers:
            layer.train()

    def to(self, device):
        pass


import os

from .base import CkptSource
from ..format import CkptFormat


class LocalCkptSource(CkptSource):

    def get(self, name: str, format: CkptFormat, prefix=None, **kwargs):
        path = name if prefix is None else os.path.join(prefix, name)
        return format.load_ckpt(path, **kwargs)

    def put(self, name: str, data, format: CkptFormat, prefix=None):
        path = name if prefix is None else os.path.join(prefix, name)
        return format.save_ckpt(data, path)



from functools import partial
from typing import Union

import torch
from tqdm import tqdm

from rainbowneko.data import CacheableDataset, BaseDataset
from rainbowneko.data import get_sampler
from rainbowneko.data.handler import DataHandler
from .base import BasicAction


class HandlerAction(BasicAction):
    def __init__(self, handler: DataHandler, key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.handler = handler

    def forward(self, **states):
        return self.handler(states)


class DataLoaderAction(BasicAction):
    def __init__(self, dataset: Union[partial, BaseDataset], actions: BasicAction, workers=0, key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.data_builder = dataset
        self.actions = actions
        self.workers = workers

    def build_dataset(self, data_builder: partial):
        batch_size = data_builder.keywords.pop("batch_size")

        dataset = data_builder()
        dataset.build_bucket(bs=batch_size, world_size=self.world_size)
        if isinstance(dataset, CacheableDataset):
            raise ValueError('Do not use CacheableDataset in workflow! Can not cache datas in workflow!')
        print(f"len(dataset): {len(dataset)}")

        return dataset, batch_size

    def build_data(self, data_builder: partial, train=False) -> torch.utils.data.DataLoader:
        drop_last = False
        dataset, batch_size = self.build_dataset(data_builder)

        # Pytorch Data loader
        sampler = get_sampler(train)(
            dataset,
            num_replicas=self.world_size,
            rank=self.local_rank,
            shuffle=train and dataset.bucket.can_shuffle,
        )
        loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=self.workers,
            sampler=sampler,
            collate_fn=dataset.collate_fn,
            drop_last=drop_last,
        )
        return loader

    def to_dev(self, x, device, dtype):
        if isinstance(x, torch.Tensor):
            if torch.is_floating_point(x):
                return x.to(device, dtype=dtype)
            else:
                return x.to(device)
        else:
            return x

    def forward(self, world_size, local_rank, device, dtype, **states):
        if not hasattr(self, 'loader'):
            self.world_size = world_size
            self.local_rank = local_rank
            self.loader = self.build_data(self.data_builder)

        states.update({"world_size": world_size, "local_rank": local_rank, "device": device, "dtype": dtype})
        for data in tqdm(self.loader):
            input_datas = {k: self.to_dev(v, device, dtype) for k, v in data.items() if k != "plugin_input"}
            if "plugin_input" in data:
                input_datas["plugin_input"] = {k: self.to_dev(v, device, dtype) for k, v in data["plugin_input"].items()}

            states_in = {**states, **input_datas}
            states = self.actions(**states_in)

        return states



import random

import numpy as np
import torch

py_state_prev = [random.getstate()]
np_state_prev = [np.random.get_state()]
torch_state_prev = [torch.get_rng_state()]
if torch.cuda.is_available():
    cuda_state_prev = [torch.cuda.get_rng_state()]


class RandomContext:
    def __init__(self, cuda=False):
        self.py_state = py_state_prev[0]
        self.np_state = np_state_prev[0]
        self.torch_state = torch_state_prev[0]

        cuda = cuda and torch.cuda.is_available()
        if cuda:
            self.cuda_state = cuda_state_prev[0]

        self.py_state_save = None
        self.np_state_save = None
        self.torch_state_save = None
        if cuda:
            self.cuda_state_save = None

        self.cuda = cuda

    def __enter__(self):
        # Save Python random state
        self.py_state_save = random.getstate()
        random.setstate(self.py_state)

        # Save NumPy random state
        self.np_state_save = np.random.get_state()
        np.random.set_state(self.np_state)

        # Save PyTorch random state
        self.torch_state_save = torch.get_rng_state()
        torch.set_rng_state(self.torch_state_save)

        # Save CUDA random state if available
        if self.cuda and torch.cuda.is_available():
            self.cuda_state_save = torch.cuda.get_rng_state()
            torch.cuda.set_rng_state(self.cuda_state)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Restore Python random state
        py_state_prev[0] = random.getstate()
        random.setstate(self.py_state_save)

        # Restore NumPy random state
        np_state_prev[0] = np.random.get_state()
        np.random.set_state(self.np_state_save)

        # Restore PyTorch random state
        torch_state_prev[0] = torch.get_rng_state()
        torch.set_rng_state(self.torch_state_save)

        # Restore CUDA random state if available
        if self.cuda and torch.cuda.is_available() and self.cuda_state is not None:
            cuda_state_prev[0] = torch.cuda.get_rng_state()
            torch.cuda.set_rng_state(self.cuda_state_save)



import argparse

from accelerate import Accelerator
from rainbowneko import _share

from .trainer_ac import Trainer, load_config_with_cli, set_seed


class TrainerSingleCard(Trainer):
    def init_context(self, cfgs_raw):
        self.accelerator = Accelerator(
            gradient_accumulation_steps=self.cfgs.train.gradient_accumulation_steps,
            mixed_precision=self.cfgs.mixed_precision,
            step_scheduler_with_optimizer=False,
        )

        self.local_rank = 0
        self.world_size = self.accelerator.num_processes
        _share.local_rank = self.local_rank
        _share.world_size = self.world_size
        _share.device = self.device

        set_seed(self.cfgs.seed + self.local_rank)

    @property
    def model_raw(self):
        return self.model_wrapper

    def boardcast_main(self, data):
        return data

    def all_gather(self, data):
        return [data]

def neko_train():
    import subprocess
    parser = argparse.ArgumentParser(description='RainbowNeko Launcher')
    parser.add_argument('--launch_cfg', type=str, default='cfgs/launcher/single.yaml')
    args, train_args = parser.parse_known_args()

    subprocess.run(["accelerate", "launch", '--config_file', args.launch_cfg, "-m",
                    "rainbowneko.train.trainer.trainer_ac_single"] + train_args, check=True)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='RainbowNeko Trainer for one GPU')
    parser.add_argument('--cfg', type=str, default='cfg/train/demo.yaml')
    args, cfg_args = parser.parse_known_args()

    parser, conf = load_config_with_cli(args.cfg, args_list=cfg_args)  # skip --cfg
    trainer = TrainerSingleCard(parser, conf)
    trainer.train()



import os
from typing import Dict, Any

from rainbowneko.utils.utils import is_image_file

from .base import VisionDataSource


class UnLabelSource(VisionDataSource):
    def __init__(self, img_root, repeat=1, **kwargs):
        super(UnLabelSource, self).__init__(img_root, repeat=repeat)

        self.img_ids = self._load_img_ids(img_root)

    def _load_img_ids(self, img_root):
        return [x for x in os.listdir(img_root) if is_image_file(x)] * self.repeat

    def __len__(self):
        return len(self.img_ids)

    def __getitem__(self, index) -> Dict[str, Any]:
        img_id = self.img_ids[index]
        path = os.path.join(self.img_root, img_id)
        return {
            'id': img_id,
            'image': path,
        }



from .pkl import PKLFormat
from typing import Dict, Any, Union

import torch
from torch.serialization import FILE_LIKE
from safetensors import safe_open
from safetensors.torch import save_file
from .base import CkptFormat

class SafeTensorFormat(CkptFormat):
    EXT = 'safetensors'

    def save_ckpt(self, sd_model: Dict[str, Any], save_f: FILE_LIKE):
        sd_unfold = self.unfold_dict(sd_model)
        save_file(sd_unfold, save_f)

    def load_ckpt(self, ckpt_f: FILE_LIKE, map_location='cpu'):
        with safe_open(ckpt_f, framework="pt", device=map_location) as f:
            sd_fold = self.fold_dict(f)
        return sd_fold

    @staticmethod
    def unfold_dict(data, split_key=':'):
        dict_unfold={}

        def unfold(prefix, dict_fold):
            for k,v in dict_fold.items():
                k_new = k if prefix=='' else f'{prefix}{split_key}{k}'
                if isinstance(v, dict):
                    unfold(k_new, v)
                elif isinstance(v, list) or isinstance(v, tuple):
                    unfold(k_new, {i:d for i,d in enumerate(v)})
                else:
                    dict_unfold[k_new]=v

        unfold('', data)
        return dict_unfold

    @staticmethod
    def fold_dict(safe_f, split_key=':'):
        dict_fold = {}

        for k in safe_f.keys():
            k_list = k.split(split_key)
            dict_last = dict_fold
            for item in k_list[:-1]:
                if item not in dict_last:
                    dict_last[item] = {}
                dict_last = dict_last[item]
            dict_last[k_list[-1]]=safe_f.get_tensor(k)

        return dict_fold


from io import BytesIO
from typing import Dict, Any

import lmdb
from PIL import Image

from .img_label import ImageLabelSource


class LMDBImageLabelSource(ImageLabelSource):
    def __init__(self, img_root, label_file, repeat=1, **kwargs):
        super().__init__(img_root, label_file, repeat, **kwargs)
        self.env = lmdb.open(img_root, readonly=True, lock=False)  # 打开 LMDB 数据库

    def __getitem__(self, index) -> Dict[str, Any]:
        img_id = self.img_ids[index]
        with self.env.begin(write=False) as txn:
            img_data = txn.get(img_id.encode())  # 从 LMDB 获取图像数据
        image = Image.open(BytesIO(img_data))
        return {
            'id': img_id,
            'image': image,
            'label': self.label_dict.get(self.img_ids[index], None)
        }



from typing import Dict, Any

from .bucket import *
from .cache import CacheableDataset, DataCache
from .dataset import BaseDataset
from .handler import *
from .label_loader import JsonLabelLoader, YamlLabelLoader, TXTLabelLoader, auto_label_loader
from .sampler import DistributedCycleSampler, get_sampler
from .source import *
from .utils import CycleData

try:
    from .webds import WebDataset
except (ImportError, ModuleNotFoundError):
    print('INFO: webdataset is not available')

class DataGroup:
    def __init__(self, loader_dict: Dict[str, Any], loss_weights: Dict[str, float], cycle=True):
        self.loader_dict = loader_dict
        self.loss_weights = loss_weights
        self.cycle = cycle

    def __iter__(self):
        if self.cycle:
            self.data_iter_dict = {name: iter(CycleData(loader)) for name, loader in self.loader_dict.items()}
        else:
            self.data_iter_dict = {}
            for name, loader in self.loader_dict.items():
                loader.dataset.bucket.rest(0)  # rest bucket
                self.data_iter_dict[name] = iter(loader)
        return self

    def __next__(self):
        if self.cycle:
            return {name: next(data_iter) for name, data_iter in self.data_iter_dict.items()}
        else:
            data_dict = {}
            for name, data_iter in self.data_iter_dict.items():
                try:
                    data_dict[name] = next(data_iter)
                except StopIteration:
                    pass
            if len(data_dict) == 0:
                raise StopIteration()
            return data_dict

    def __len__(self):
        return max([len(loader) for loader in self.loader_dict.values()])

    def get_dataset(self, name):
        return self.loader_dict[name].dataset

    def get_loss_weights(self, name):
        return self.loss_weights[name]

    def first_loader(self):
        return next(iter(self.loader_dict.values()))



"""
plugin.py
====================
    :Name:        model plugin
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     10/03/2023
    :Licence:     Apache-2.0
"""

import weakref
import re
from typing import Tuple, List, Dict, Any, Iterable, Callable, Optional

import torch
from torch import nn

from rainbowneko.utils import split_module_name


class BasePluginBlock(nn.Module):
    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def forward(self, host: nn.Module, fea_in: Tuple[torch.Tensor], fea_out: torch.Tensor):
        return fea_out

    def register_forward_hook_to(self, host: nn.Module,
                                 hook: Callable[[nn.Module, Tuple[Any, ...], Dict[str, Any], Any], Optional[Any]]):
        try:
            return host.register_forward_hook(hook, with_kwargs=True)
        except TypeError:  # torch < 2.0
            return host.register_forward_hook(lambda host, args, output: hook(host, args, {}, output))

    def register_forward_pre_hook_to(self, host: nn.Module, hook: Callable[
        [nn.Module, Tuple[Any, ...], Dict[str, Any]], Optional[Tuple[Any, Dict[str, Any]]]]):
        try:
            return host.register_forward_pre_hook(hook, with_kwargs=True)
        except TypeError:  # torch < 2.0
            return host.register_forward_pre_hook(lambda host, args: hook(host, args, {})[0])

    def remove(self):
        pass

    def feed_input_data(self, data):
        self.input_data = data

    def register_input_feeder_to(self, host_model):
        if not hasattr(host_model, 'input_feeder'):
            host_model.input_feeder = []
        host_model.input_feeder.append(self.feed_input_data)

    def set_hyper_params(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def extract_state_without_plugin(model: nn.Module, trainable=False):
        trainable_keys = {k for k, v in model.named_parameters() if v.requires_grad}
        plugin_names = {k for k, v in model.named_modules() if isinstance(v, BasePluginBlock)}
        model_sd = {}
        for k, v in model.state_dict().items():
            if (not trainable) or k in trainable_keys:
                for name in plugin_names:
                    if k.startswith(name):
                        break
                else:
                    model_sd[k] = v
        return model_sd

    def get_trainable_parameters(self) -> Iterable[nn.Parameter]:
        return self.parameters()


class WrapablePlugin:
    wrapable_classes = ()

    @classmethod
    def wrap_layer(cls, name: str, layer: nn.Module, **kwargs):
        plugin = cls(name, layer, **kwargs)
        return plugin

    @classmethod
    def named_modules_with_exclude(cls, self, memo=None, prefix: str = '', remove_duplicate: bool = True,
                                   exclude_key=None, exclude_classes=tuple()):

        if memo is None:
            memo = set()
        if self not in memo:
            if remove_duplicate:
                memo.add(self)
            if (exclude_key is None or not re.search(exclude_key, prefix)) and not isinstance(self, exclude_classes):
                yield prefix, self
                for name, module in self._modules.items():
                    if module is None:
                        continue
                    submodule_prefix = prefix + ('.' if prefix else '') + name
                    for m in cls.named_modules_with_exclude(module, memo, submodule_prefix, remove_duplicate, exclude_key,
                                                            exclude_classes):
                        yield m

    @classmethod
    def wrap_model(cls, name: str, host: nn.Module, exclude_key=None, exclude_classes=tuple(),
                   **kwargs):  # -> Dict[str, SinglePluginBlock]:
        '''
        parent_block and other args required in __init__ will be put into kwargs, compatible with multiple models.
        '''
        plugin_block_dict = {}
        if isinstance(host, cls.wrapable_classes):
            plugin_block_dict[''] = cls.wrap_layer(name, host, **kwargs)
        else:
            named_modules = {layer_name: layer for layer_name, layer in cls.named_modules_with_exclude(
                host, exclude_key=exclude_key, exclude_classes=exclude_classes)}
            for layer_name, layer in named_modules.items():
                if isinstance(layer, cls.wrapable_classes):
                    # For plugins that need parent_block
                    if 'parent_block' in kwargs:
                        parent_name, host_name = split_module_name(layer_name)
                        kwargs['parent_block'] = named_modules[parent_name]
                        kwargs['host_name'] = host_name
                    plugin_block_dict[layer_name] = cls.wrap_layer(name, layer, **kwargs)
        return plugin_block_dict


class SinglePluginBlock(BasePluginBlock, WrapablePlugin):

    def __init__(self, name: str, host: nn.Module, hook_param=None, host_model=None):
        super().__init__(name)
        self.host = weakref.ref(host)
        setattr(host, name, self)

        if hook_param is None:
            self.hook_handle = self.register_forward_hook_to(host, self.layer_hook)
        else:  # hook for model parameters
            self.backup = getattr(host, hook_param)
            self.target = hook_param
            self.handle_pre = self.register_forward_pre_hook_to(host, self.pre_hook)
            self.handle_post = self.register_forward_hook_to(host, self.post_hook)

    def layer_hook(self, host: nn.Module, args: Tuple[Any, ...], kwargs: Dict[str, Any], feat_out: Any):
        return self(feat_out, *args, **kwargs)

    def pre_hook(self, host: nn.Module, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        host.weight_restored = False
        host_param = getattr(host, self.target)
        delattr(host, self.target)
        setattr(host, self.target, self(host_param))

    def post_hook(self, host, args, kwargs, feat_out):
        if not getattr(host, 'weight_restored', False):
            setattr(host, self.target, self.backup)
            host.weight_restored = True

    def remove(self):
        host = self.host()
        delattr(host, self.name)
        if hasattr(self, 'hook_handle'):
            self.hook_handle.remove()
        else:
            self.handle_pre.remove()
            self.handle_post.remove()


class PluginBlock(BasePluginBlock):
    def __init__(self, name, from_layer: Dict[str, Any], to_layer: Dict[str, Any], host_model=None):
        super().__init__(name)
        self.host_from = weakref.ref(from_layer['layer'])
        self.host_to = weakref.ref(to_layer['layer'])
        setattr(from_layer['layer'], name, self)

        if from_layer['pre_hook']:
            self.hook_handle_from = self.register_forward_pre_hook_to(from_layer['layer'], self.from_layer_hook)
        else:
            self.hook_handle_from = self.register_forward_hook_to(from_layer['layer'], self.from_layer_hook)

        if to_layer['pre_hook']:
            self.hook_handle_to = self.register_forward_pre_hook_to(to_layer['layer'], self.to_layer_hook)
        else:
            self.hook_handle_to = self.register_forward_hook_to(to_layer['layer'], self.to_layer_hook)

    def from_layer_hook(self, host, args: Tuple[Any, ...], kwargs: Dict[str, Any], fea_out: Any=None):
        self.feat_from = (args, kwargs)

    def to_layer_hook(self, host, args: Tuple[Any, ...], kwargs: Dict[str, Any], fea_out: Any=None):
        return self(self.feat_from[0], self.feat_from[1], fea_out, *args, **kwargs)

    def remove(self):
        host_from = self.host_from()
        delattr(host_from, self.name)
        self.hook_handle_from.remove()
        self.hook_handle_to.remove()


class MultiPluginBlock(BasePluginBlock):
    def __init__(self, name: str, from_layers: List[Dict[str, Any]], to_layers: List[Dict[str, Any]], host_model=None):
        super().__init__(name)
        assert host_model is not None
        self.host_from = [weakref.ref(x['layer']) for x in from_layers]
        self.host_to = [weakref.ref(x['layer']) for x in to_layers]
        self.host_model = weakref.ref(host_model)
        setattr(host_model, name, self)

        self.feat_from = [None for _ in range(len(from_layers))]

        self.hook_handle_from = []
        self.hook_handle_to = []

        for idx, layer in enumerate(from_layers):
            if layer['pre_hook']:
                handle_from = self.register_forward_pre_hook_to(layer['layer'],
                        lambda host, args, kwargs, idx=idx: self.from_layer_hook(host, idx, args, kwargs))
            else:
                handle_from = self.register_forward_hook_to(layer['layer'],
                        lambda host, args, kwargs, fea_out, idx=idx: self.from_layer_hook(host, idx, args, kwargs, fea_out))
            self.hook_handle_from.append(handle_from)
        for idx, layer in enumerate(to_layers):
            if layer['pre_hook']:
                handle_to = self.register_forward_pre_hook_to(layer['layer'],
                        lambda host, args, kwargs, idx=idx: self.to_layer_hook(host, idx, args, kwargs))
            else:
                handle_to = self.register_forward_hook_to(layer['layer'],
                        lambda host, args, kwargs, fea_out, idx=idx: self.to_layer_hook(host, idx, args, kwargs, fea_out))
            self.hook_handle_to.append(handle_to)

        self.record_count = 0

    def from_layer_hook(self, host, idx: int, args: Tuple[Any, ...], kwargs: Dict[str, Any], fea_out: Any=None):
        self.feat_from[idx] = (args, kwargs)
        self.record_count += 1
        if self.record_count == len(self.feat_from):  # call forward when all feat is record
            self.record_count = 0
            self.feat_to = self(self.feat_from)

    def to_layer_hook(self, host, idx: int, args: Tuple[Any, ...], kwargs: Dict[str, Any], fea_out: Any=None):
        return self.feat_to[idx] + fea_out

    def remove(self):
        host_model = self.host_model()
        delattr(host_model, self.name)
        for handle_from in self.hook_handle_from:
            handle_from.remove()
        for handle_to in self.hook_handle_to:
            handle_to.remove()


class PatchPluginContainer(nn.Module):
    def __init__(self, host_name, host, parent_block):
        super().__init__()
        self._host = host
        self.host_name = host_name
        self.parent_block = weakref.ref(parent_block)
        self.plugin_names = []

        delattr(parent_block, host_name)
        setattr(parent_block, host_name, self)

    def add_plugin(self, name: str, plugin: 'PatchPluginBlock'):
        setattr(self, name, plugin)
        self.plugin_names.append(name)

    def remove_plugin(self, name: str):
        delattr(self, name)
        self.plugin_names.remove(name)
        if len(self.plugin_names) == 0:
            self.remove()

    def forward(self, *args, **kwargs):
        for name, plugin in self:
            args, kwargs = plugin.pre_forward(*args, **kwargs)
        output = self._host(*args, **kwargs)
        for name, plugin in self:
            output = plugin.post_forward(output, *args, **kwargs)
        return output

    def remove(self):
        parent_block = self.parent_block()
        delattr(parent_block, self.host_name)
        setattr(parent_block, self.host_name, self._host)

    def __iter__(self):
        for name in self.plugin_names:
            yield name, self[name]

    def __getitem__(self, name):
        return getattr(self, name)


class PatchPluginBlock(BasePluginBlock, WrapablePlugin):
    container_cls = PatchPluginContainer

    def __init__(self, name: str, host: nn.Module, host_model=None, parent_block: nn.Module = None, host_name: str = None):
        super().__init__(name)
        if isinstance(host, self.container_cls):
            self.host = weakref.ref(host._host)
        else:
            self.host = weakref.ref(host)
        self.parent_block = weakref.ref(parent_block)
        self.host_name = host_name

        container = self.get_container(host, host_name, parent_block)
        container.add_plugin(name, self)
        self.container = weakref.ref(container)

    def pre_forward(self, *args, **kwargs):
        return args, kwargs

    def post_forward(self, output, *args, **kwargs):
        return output

    def remove(self):
        container = self.container()
        container.remove_plugin(self.name)

    def get_container(self, host, host_name, parent_block):
        if isinstance(host, self.container_cls):
            return host
        else:
            return self.container_cls(host_name, host, parent_block)

    @classmethod
    def wrap_model(cls, name: str, host: nn.Module, exclude_key=None, exclude_classes=tuple(),
                   **kwargs):  # -> Dict[str, SinglePluginBlock]:
        '''
        parent_block and other args required in __init__ will be put into kwargs, compatible with multiple models.
        '''
        plugin_block_dict = {}
        if isinstance(host, cls.wrapable_classes):
            plugin_block_dict[''] = cls.wrap_layer(name, host, **kwargs)
        else:
            named_modules = {layer_name: layer for layer_name, layer in cls.named_modules_with_exclude(
                host, exclude_key=exclude_key or '_host', exclude_classes=exclude_classes)}
            for layer_name, layer in named_modules.items():
                if isinstance(layer, cls.wrapable_classes) or isinstance(layer, cls.container_cls):
                    # For plugins that need parent_block
                    if kwargs.get('parent_block', False):
                        parent_name, host_name = split_module_name(layer_name)
                        kwargs['parent_block'] = named_modules[parent_name]
                        kwargs['host_name'] = host_name
                    plugin_block_dict[layer_name] = cls.wrap_layer(name, layer, **kwargs)
        return plugin_block_dict


class PluginGroup:
    def __init__(self, plugin_dict: Dict[str, BasePluginBlock]):
        self.plugin_dict = plugin_dict  # {host_model_path: plugin_object}

    def __setitem__(self, k, v):
        self.plugin_dict[k] = v

    def __getitem__(self, k):
        return self.plugin_dict[k]

    @property
    def plugin_name(self):
        if self.empty():
            return None
        return next(iter(self.plugin_dict.values())).name

    def remove(self):
        for plugin in self.plugin_dict.values():
            plugin.remove()

    def state_dict(self, model=None):
        if model is None:
            return {f'{k}.___.{ks}': vs for k, v in self.plugin_dict.items() for ks, vs in v.state_dict().items()}
        else:
            sd_model = model.state_dict()
            return {f'{k}.___.{ks}': sd_model[f'{k}.{v.name}.{ks}'] for k, v in self.plugin_dict.items() for ks, vs in
                    v.state_dict().items()}

    def from_model(self, model):
        named_module = {name: layer for name, layer in model.named_modules()}
        return PluginGroup({k: named_module[f'{k}.{v.name}'] for k, v in self.plugin_dict.items()})

    def empty(self):
        return len(self.plugin_dict) == 0



from .base import DataHandler, HandlerGroup, HandlerChain, SyncHandler
from .image import ImageHandler, LoadImageHandler, AutoSizeHandler
from .label import LabelHandler
from .batch import MixUPHandler
from .frame import FrameHandler, FrameCatHandler


from typing import Any, Dict

import numpy as np


class BaseBucket:
    can_shuffle = True

    def __getitem__(self, idx) -> Dict[str, Any]:
        '''
        :return: (file name of image), (target image size)
        '''
        return self.source[idx]

    def _shuffle(self, bufsize=1000, initial=100, rs: np.random.RandomState = None):
        """Shuffle the data in the stream.
        Modify from webdataset
        """
        source_iter = iter(self.source)

        initial = min(initial, bufsize)
        buf = []

        def pick():
            k = rs.randint(0, len(buf) - 1)
            sample = buf[k]
            buf[k] = buf[-1]
            buf.pop()
            return sample

        for datas in source_iter:
            buf.append(datas)
            if len(buf) < bufsize:
                try:
                    buf.append(next(source_iter))
                except StopIteration:
                    pass
            if len(buf) >= initial:
                yield pick()
        while len(buf) > 0:
            yield pick()

    def next_data(self, shuffle=True):
        if not hasattr(self, 'buffer_iter'):
            self.buffer_iter = self._shuffle(rs=self.rs) if shuffle else iter(self.source)
        return next(self.buffer_iter)

    def __len__(self):
        return len(self.source)

    def build(self, bs: int, world_size: int, source: 'DataSource'):
        self.source = source

        try:
            _ = self.source[0]
            self.source_indexable = True
        except NotImplementedError:
            self.source_indexable = False

    def rest(self, epoch):
        if not self.source_indexable:
            self.rs = np.random.RandomState(42 + epoch)



from pathlib import Path
from typing import Any, Dict, Union

from rainbowneko.data.label_loader import BaseLabelLoader
from rainbowneko.utils import Path_Like

from .img_label import ImageLabelSource


class ImageFolderClassSource(ImageLabelSource):
    def __init__(self, img_root, repeat=1, **kwargs):
        super().__init__(img_root, img_root, repeat=repeat, **kwargs)

    def __getitem__(self, index) -> Dict[str, Any]:
        img_id = self.img_ids[index]
        path = self.img_root / img_id
        return {
            'id': img_id,
            'image': path,
            'label': self.cls_id_dict[self.label_dict[self.img_ids[index]]]
        }

    def _load_label_data(self, img_root: Union[Path_Like, BaseLabelLoader]):
        ''' {class_name/image.ext: label} '''
        if img_root is None:
            label_dict = {}
        elif isinstance(img_root, Path_Like):
            label_dict = {}
            img_root = Path(img_root)
            for class_folder in img_root.iterdir():
                cls_name = class_folder.name
                for img_path in class_folder.iterdir():
                    label_dict[f'{cls_name}/{img_path.name}'] = cls_name
        else:
            label_dict = img_root.load()

        # label to cls_id
        self.cls_id_dict = {cls_name: cls_id for cls_id, cls_name in enumerate(sorted(set(label_dict.values())))}
        return label_dict



# coding=utf-8
# Copyright 2023 The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LR Scheduler copy from diffusers."""

import math
from enum import Enum
from functools import partial
from typing import Optional, Union, Callable

from torch.optim import Optimizer
from torch.optim import lr_scheduler
from torch.optim.lr_scheduler import LambdaLR

SchedulerType = Callable[[int], float]


class SchedulerName(Enum):
    LINEAR = "linear"
    COSINE = "cosine"
    COSINE_WITH_RESTARTS = "cosine_with_restarts"
    POLYNOMIAL = "polynomial"
    CONSTANT = "constant"
    CONSTANT_WITH_WARMUP = "constant_with_warmup"
    PIECEWISE_CONSTANT = "piecewise_constant"


def constant_schedule() -> SchedulerType:
    """
    Create a schedule with a constant learning rate, using the learning rate set in optimizer.

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """
    return lambda _: 1


def constant_schedule_with_warmup(num_warmup_steps: int) -> SchedulerType:
    """
    Create a schedule with a constant learning rate preceded by a warmup period during which the learning rate
    increases linearly between 0 and the initial lr set in the optimizer.

    Args:
        num_warmup_steps (`int`):
            The number of steps for the warmup phase.

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1.0, num_warmup_steps))
        return 1.0

    return lr_lambda


def piecewise_constant_schedule(step_rules: str) -> SchedulerType:
    """
    Create a schedule with a constant learning rate, using the learning rate set in optimizer.

    Args:
        step_rules (`string`):
            The rules for the learning rate. ex: rule_steps="1:10,0.1:20,0.01:30,0.005" it means that the learning rate
            if multiple 1 for the first 10 steps, mutiple 0.1 for the next 20 steps, multiple 0.01 for the next 30
            steps and multiple 0.005 for the other steps.

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    rules_dict = {}
    rule_list = step_rules.split(",")
    for rule_str in rule_list[:-1]:
        value_str, steps_str = rule_str.split(":")
        steps = int(steps_str)
        value = float(value_str)
        rules_dict[steps] = value
    last_lr_multiple = float(rule_list[-1])

    def create_rules_function(rules_dict, last_lr_multiple):
        def rule_func(steps: int) -> float:
            sorted_steps = sorted(rules_dict.keys())
            for i, sorted_step in enumerate(sorted_steps):
                if steps < sorted_step:
                    return rules_dict[sorted_steps[i]]
            return last_lr_multiple

        return rule_func

    rules_func = create_rules_function(rules_dict, last_lr_multiple)

    return rules_func


def linear_schedule_with_warmup(num_warmup_steps, num_training_steps) -> SchedulerType:
    """
    Create a schedule with a learning rate that decreases linearly from the initial lr set in the optimizer to 0, after
    a warmup period during which it increases linearly from 0 to the initial lr set in the optimizer.

    Args:
        num_warmup_steps (`int`):
            The number of steps for the warmup phase.
        num_training_steps (`int`):
            The total number of training steps.

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    return lr_lambda


def cosine_schedule_with_warmup(num_warmup_steps: int, num_training_steps: int, num_cycles: float = 0.5):
    """
    Create a schedule with a learning rate that decreases following the values of the cosine function between the
    initial lr set in the optimizer to 0, after a warmup period during which it increases linearly between 0 and the
    initial lr set in the optimizer.

    Args:
        num_warmup_steps (`int`):
            The number of steps for the warmup phase.
        num_training_steps (`int`):
            The total number of training steps.
        num_cycles (`float`, *optional*, defaults to 0.5):
            The number of periods of the cosine function in a schedule (the default is to just decrease from the max
            value to 0 following a half-cosine).

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    return lr_lambda


def cosine_with_hard_restarts_schedule_with_warmup(num_warmup_steps: int, num_training_steps: int, num_cycles: int = 1):
    """
    Create a schedule with a learning rate that decreases following the values of the cosine function between the
    initial lr set in the optimizer to 0, with several hard restarts, after a warmup period during which it increases
    linearly between 0 and the initial lr set in the optimizer.

    Args:
        num_warmup_steps (`int`):
            The number of steps for the warmup phase.
        num_training_steps (`int`):
            The total number of training steps.
        num_cycles (`int`, *optional*, defaults to 1):
            The number of hard restarts to use.

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        if progress >= 1.0:
            return 0.0
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * ((float(num_cycles) * progress) % 1.0))))

    return lr_lambda


def polynomial_decay_schedule_with_warmup(num_warmup_steps, num_training_steps, v_start, v_end=1e-7, power=1.0):
    """
    Create a schedule with a learning rate that decreases as a polynomial decay from the initial lr set in the
    optimizer to end lr defined by *lr_end*, after a warmup period during which it increases linearly from 0 to the
    initial lr set in the optimizer.

    Args:
        num_warmup_steps (`int`):
            The number of steps for the warmup phase.
        num_training_steps (`int`):
            The total number of training steps.
        lr_end (`float`, *optional*, defaults to 1e-7):
            The end LR.
        power (`float`, *optional*, defaults to 1.0):
            Power factor.

    Note: *power* defaults to 1.0 as in the fairseq implementation, which in turn is based on the original BERT
    implementation at
    https://github.com/google-research/bert/blob/f39e881b169b9d53bea03d2d341b31707a6c052b/optimization.py#L37

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.

    """

    if not (v_start > v_end):
        raise ValueError(f"v_end ({v_end}) must be be smaller than v_start ({v_start})")

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        elif current_step > num_training_steps:
            return v_end / v_start  # as LambdaLR multiplies by lr_init
        else:
            lr_range = v_start - v_end
            decay_steps = num_training_steps - num_warmup_steps
            pct_remaining = 1 - (current_step - num_warmup_steps) / decay_steps
            decay = lr_range * pct_remaining ** power + v_end
            return decay / v_start  # as LambdaLR multiplies by lr_init

    return lr_lambda


def fractional_warmup_schedule(decay_max=0.997, decay_factor=(1, 10)):
    def lr_lambda(current_step: int):
        value = (decay_factor[0] + current_step) / (decay_factor[1] + current_step)
        return min(decay_max, value)

    return lr_lambda


def polynomial_EMA_schedule(decay_max=0.997, inv_gamma=1.0, power=3 / 4):
    def lr_lambda(current_step: int):
        value = 1 - (1 + current_step / inv_gamma) ** -power
        return min(decay_max, value)

    return lr_lambda


TYPE_TO_SCHEDULER_FUNCTION = {
    SchedulerName.LINEAR: linear_schedule_with_warmup,
    SchedulerName.COSINE: cosine_schedule_with_warmup,
    SchedulerName.COSINE_WITH_RESTARTS: cosine_with_hard_restarts_schedule_with_warmup,
    SchedulerName.POLYNOMIAL: polynomial_decay_schedule_with_warmup,
    SchedulerName.CONSTANT: constant_schedule,
    SchedulerName.CONSTANT_WITH_WARMUP: constant_schedule_with_warmup,
    SchedulerName.PIECEWISE_CONSTANT: piecewise_constant_schedule,
}


def get_scheduler_with_name(
        name: Union[str, SchedulerName],
        optimizer: Optimizer,
        warmup_steps: Optional[int] = None,
        training_steps: Optional[int] = None,
        **kwargs
):
    """
    Unified API to get any scheduler from its name.

    Args:
        name (`str` or `SchedulerName`):
            The name of the scheduler to use.
        optimizer (`torch.optim.Optimizer`):
            The optimizer that will be used during training.
        warmup_steps (`int`, *optional*):
            The number of warmup steps to do. This is not required by all schedulers (hence the argument being
            optional), the function will raise an error if it's unset and the scheduler type requires it.
        training_steps (`int``, *optional*):
            The number of training steps to do. This is not required by all schedulers (hence the argument being
            optional), the function will raise an error if it's unset and the scheduler type requires it.
        num_cycles (`int`, *optional*):
            The number of hard restarts used in `COSINE_WITH_RESTARTS` scheduler.
        power (`float`, *optional*, defaults to 1.0):
            Power factor. See `POLYNOMIAL` scheduler
        last_epoch (`int`, *optional*, defaults to -1):
            The index of the last epoch when resuming training.
    """
    # All other schedulers require `num_training_steps`
    if training_steps is None:
        raise ValueError(f"{name} requires `num_training_steps`, please provide that argument.")

    # All other schedulers require `num_warmup_steps`
    if warmup_steps is None:
        raise ValueError(f"{name} requires `num_warmup_steps`, please provide that argument.")

    if isinstance(warmup_steps, float):  # warmup ratio
        warmup_steps = int(warmup_steps * training_steps)

    # One Cycle for super convergence
    if name == 'one_cycle':
        scheduler = lr_scheduler.OneCycleLR(optimizer, max_lr=[x['lr'] for x in optimizer.state_dict()['param_groups']],
                                            steps_per_epoch=training_steps, epochs=1,
                                            pct_start=warmup_steps / training_steps, **kwargs)
        return scheduler

    name = SchedulerName(name)
    schedule_func = TYPE_TO_SCHEDULER_FUNCTION[name]
    if name == SchedulerName.CONSTANT:
        scheduler = schedule_func()
    elif name == SchedulerName.CONSTANT_WITH_WARMUP:
        scheduler = schedule_func(num_warmup_steps=warmup_steps)
    elif name == SchedulerName.COSINE_WITH_RESTARTS:
        scheduler = schedule_func(num_warmup_steps=warmup_steps, num_training_steps=training_steps, **kwargs)
    elif name == SchedulerName.POLYNOMIAL:
        scheduler = schedule_func(num_warmup_steps=warmup_steps, num_training_steps=training_steps, **kwargs)
    else:
        scheduler = schedule_func(num_warmup_steps=warmup_steps, num_training_steps=training_steps, **kwargs)
    return LambdaLR(optimizer, scheduler)


def get_scheduler(cfg, optimizer, num_training_steps):
    if cfg is None:
        return None
    elif isinstance(cfg, partial):
        try:
            return cfg(optimizer=optimizer, training_steps=num_training_steps)
        except:
            return cfg(optimizer=optimizer)
    else:
        return get_scheduler_with_name(optimizer=optimizer, training_steps=num_training_steps, **cfg)


def ConstantLR(optimizer: Optimizer, warmup_steps: int = None, last_epoch: int = -1):
    if warmup_steps is None:
        scheduler = constant_schedule()
    else:
        scheduler = constant_schedule_with_warmup(warmup_steps)
    return LambdaLR(optimizer, scheduler, last_epoch=last_epoch)


def MultiStepLR(optimizer: Optimizer, step_rules: str, last_epoch: int = -1):
    scheduler = piecewise_constant_schedule(step_rules)
    return LambdaLR(optimizer, scheduler, last_epoch=last_epoch)


def CosineLR(optimizer: Optimizer, training_steps: int, warmup_steps: int = 0, num_cycles: float = 0.5, last_epoch: int = -1):
    scheduler = cosine_schedule_with_warmup(warmup_steps, training_steps, num_cycles)
    return LambdaLR(optimizer, scheduler, last_epoch=last_epoch)


def CosineRestartLR(optimizer: Optimizer, training_steps: int, warmup_steps: int = 0, num_cycles: int = 1, last_epoch: int = -1):
    scheduler = cosine_with_hard_restarts_schedule_with_warmup(warmup_steps, training_steps, num_cycles)
    return LambdaLR(optimizer, scheduler, last_epoch=last_epoch)


def PolynomialLR(optimizer, training_steps: int, warmup_steps: int = 0, lr_end=1e-7, power=1.0, last_epoch=-1):
    lr_init = optimizer.defaults["lr"]
    scheduler = polynomial_decay_schedule_with_warmup(warmup_steps, training_steps, lr_init, lr_end, power)
    return LambdaLR(optimizer, scheduler, last_epoch=last_epoch)


def OneCycleLR(optimizer: Optimizer, training_steps: int, warmup_steps: int = None, **kwargs):
    return lr_scheduler.OneCycleLR(optimizer, max_lr=[x['lr'] for x in optimizer.state_dict()['param_groups']],
                                   steps_per_epoch=training_steps, epochs=1, pct_start=warmup_steps / training_steps, **kwargs)



"""

img_size_tool.py
====================

    :Name:        get_image_size
    :Purpose:     extract image dimensions given a file path

    :Author:      Paulo Scardine (based on code from Emmanuel VAÏSSE)
                  Dong Ziyi, add webp support

    :Created:     26/09/2013
    :Modified:    02/03/2023
    :Copyright:   (c) Paulo Scardine 2013
    :Licence:     MIT

"""

import collections
import os
import io
import struct
from PIL import Image

FILE_UNKNOWN = "Sorry, don't know how to get size for this file."

class UnknownImageFormat(Exception):
    pass

types_support = ['bmp', 'gif', 'ico', 'jpeg', 'jpg', 'png', 'tiff', 'webp']

def get_image_size(file_path):
    """
    Return (width, height) for a given img file content - no external
    dependencies except the os and struct builtin modules
    """
    width, height = get_image_metadata(file_path)
    return width, height


def get_image_size_from_bytesio(input, size):
    """
    Return (width, height) for a given img file content - no external
    dependencies except the os and struct builtin modules

    Args:
        input (io.IOBase): io object support read & seek
        size (int): size of buffer in byte
    """
    width, height = get_image_metadata_from_bytesio(input, size)
    return width, height


def get_image_metadata(file_path):
    """
    Return an `Image` object for a given img file content - no external
    dependencies except the os and struct builtin modules

    Args:
        file_path (str): path to an image file

    Returns:
        (width, height)
    """
    size = os.path.getsize(file_path)

    # be explicit with open arguments - we need binary mode
    with io.open(file_path, "rb") as input:
        return get_image_metadata_from_bytesio(input, size, file_path)


def get_image_metadata_from_bytesio(input, size, file_path=None):
    """
    Return an `Image` object for a given img file content - no external
    dependencies except the os and struct builtin modules

    Args:
        input (io.IOBase): io object support read & seek
        size (int): size of buffer in byte
        file_path (str): path to an image file

    Returns:
        (width, height)
    """
    height = -1
    width = -1
    data = input.read(30)
    msg = " raised while trying to decode as JPEG."

    if (size >= 10) and data[:6] in (b'GIF87a', b'GIF89a'):
        # GIFs
        #imgtype = GIF
        w, h = struct.unpack("<HH", data[6:10])
        width = int(w)
        height = int(h)
    elif (size >= 24) and data[8:12] == b'WEBP':
        # WEBPs
        #imgtype = WEBP
        if data[15]==b'X': #VP8X
            w = int.from_bytes(data[24:27], 'little')+1
            h = int.from_bytes(data[27:30], 'little')+1
        elif data[15]==b' ': #VP8
            w, h = struct.unpack("<HH", data[0x1A:0x1E])
        else:
            w, h = Image.open(file_path).size

        width = int(w)
        height = int(h)
    elif ((size >= 24) and data.startswith(b'\211PNG\r\n\032\n')
            and (data[12:16] == b'IHDR')):
        # PNGs
        #imgtype = PNG
        w, h = struct.unpack(">LL", data[16:24])
        width = int(w)
        height = int(h)
    elif (size >= 16) and data.startswith(b'\211PNG\r\n\032\n'):
        # older PNGs
        #imgtype = PNG
        w, h = struct.unpack(">LL", data[8:16])
        width = int(w)
        height = int(h)
    elif (size >= 2) and data.startswith(b'\377\330'):
        # JPEG
        #imgtype = JPEG
        input.seek(0)
        input.read(2)
        b = input.read(1)
        try:
            while (b and ord(b) != 0xDA):
                while (ord(b) != 0xFF):
                    b = input.read(1)
                while (ord(b) == 0xFF):
                    b = input.read(1)
                if (ord(b) >= 0xC0 and ord(b) <= 0xC3):
                    input.read(3)
                    h, w = struct.unpack(">HH", input.read(4))
                    break
                else:
                    input.read(
                        int(struct.unpack(">H", input.read(2))[0]) - 2)
                b = input.read(1)
            width = int(w)
            height = int(h)
        except struct.error:
            raise UnknownImageFormat("StructError" + msg)
        except ValueError:
            raise UnknownImageFormat("ValueError" + msg)
        except Exception as e:
            raise UnknownImageFormat(e.__class__.__name__ + msg)
    elif (size >= 26) and data.startswith(b'BM'):
        # BMP
        #imgtype = BMP
        headersize = struct.unpack("<I", data[14:18])[0]
        if headersize == 12:
            w, h = struct.unpack("<HH", data[18:22])
            width = int(w)
            height = int(h)
        elif headersize >= 40:
            w, h = struct.unpack("<ii", data[18:26])
            width = int(w)
            # as h is negative when stored upside down
            height = abs(int(h))
        else:
            raise UnknownImageFormat(
                "Unkown DIB header size:" +
                str(headersize))
    elif (size >= 8) and data[:4] in (b"II\052\000", b"MM\000\052"):
        # Standard TIFF, big- or little-endian
        # BigTIFF and other different but TIFF-like formats are not
        # supported currently
        #imgtype = TIFF
        byteOrder = data[:2]
        boChar = ">" if byteOrder == "MM" else "<"
        # maps TIFF type id to size (in bytes)
        # and python format char for struct
        tiffTypes = {
            1: (1, boChar + "B"),  # BYTE
            2: (1, boChar + "c"),  # ASCII
            3: (2, boChar + "H"),  # SHORT
            4: (4, boChar + "L"),  # LONG
            5: (8, boChar + "LL"),  # RATIONAL
            6: (1, boChar + "b"),  # SBYTE
            7: (1, boChar + "c"),  # UNDEFINED
            8: (2, boChar + "h"),  # SSHORT
            9: (4, boChar + "l"),  # SLONG
            10: (8, boChar + "ll"),  # SRATIONAL
            11: (4, boChar + "f"),  # FLOAT
            12: (8, boChar + "d")   # DOUBLE
        }
        ifdOffset = struct.unpack(boChar + "L", data[4:8])[0]
        try:
            countSize = 2
            input.seek(ifdOffset)
            ec = input.read(countSize)
            ifdEntryCount = struct.unpack(boChar + "H", ec)[0]
            # 2 bytes: TagId + 2 bytes: type + 4 bytes: count of values + 4
            # bytes: value offset
            ifdEntrySize = 12
            for i in range(ifdEntryCount):
                entryOffset = ifdOffset + countSize + i * ifdEntrySize
                input.seek(entryOffset)
                tag = input.read(2)
                tag = struct.unpack(boChar + "H", tag)[0]
                if(tag == 256 or tag == 257):
                    # if type indicates that value fits into 4 bytes, value
                    # offset is not an offset but value itself
                    type = input.read(2)
                    type = struct.unpack(boChar + "H", type)[0]
                    if type not in tiffTypes:
                        raise UnknownImageFormat(
                            "Unkown TIFF field type:" +
                            str(type))
                    typeSize = tiffTypes[type][0]
                    typeChar = tiffTypes[type][1]
                    input.seek(entryOffset + 8)
                    value = input.read(typeSize)
                    value = int(struct.unpack(typeChar, value)[0])
                    if tag == 256:
                        width = value
                    else:
                        height = value
                if width > -1 and height > -1:
                    break
        except Exception as e:
            raise UnknownImageFormat(str(e))
    elif size >= 2:
            # see http://en.wikipedia.org/wiki/ICO_(file_format)
        #imgtype = 'ICO'
        input.seek(0)
        reserved = input.read(2)
        if 0 != struct.unpack("<H", reserved)[0]:
            raise UnknownImageFormat(FILE_UNKNOWN)
        format = input.read(2)
        assert 1 == struct.unpack("<H", format)[0]
        num = input.read(2)
        num = struct.unpack("<H", num)[0]
        if num > 1:
            import warnings
            warnings.warn("ICO File contains more than one image")
        # http://msdn.microsoft.com/en-us/library/ms997538.aspx
        w = input.read(1)
        h = input.read(1)
        width = ord(w)
        height = ord(h)
    else:
        raise UnknownImageFormat(FILE_UNKNOWN)

    return width, height


import re
from typing import Dict, List, Union, Any

from torch import nn


def get_class_match_layer(class_name, block: nn.Module):
    if type(block).__name__ == class_name:
        return [""]
    else:
        return ["." + name for name, layer in block.named_modules() if type(layer).__name__ == class_name]


def get_match_layers(layers, all_layers, return_metas=False) -> Union[List[str], List[Dict[str, Any]]]:
    res = []
    for name in layers:
        metas = name.split(":")

        use_re = False
        pre_hook = False
        cls_filter = None
        for meta in metas[:-1]:
            if meta == "re":
                use_re = True
            elif meta == "pre_hook":
                pre_hook = True
            elif meta.startswith("cls("):
                cls_filter = meta[4:-1]

        name = metas[-1]
        if use_re:
            pattern = re.compile(name)
            match_layers = filter(lambda x: pattern.match(x) != None, all_layers.keys())
        else:
            match_layers = [name]

        if cls_filter is not None:
            match_layers_new = []
            for layer in match_layers:
                match_layers_new.extend([layer + x for x in get_class_match_layer(name[1], all_layers[layer])])
            match_layers = match_layers_new

        for layer in match_layers:
            if return_metas:
                res.append({"layer": layer, "pre_hook": pre_hook})
            else:
                res.append(layer)

    # Remove duplicates and keep the original order
    if return_metas:
        layer_set = set()
        res_unique = []
        for item in res:
            if item["layer"] not in layer_set:
                layer_set.add(item["layer"])
                res_unique.append(item)
        return res_unique
    else:
        return sorted(set(res), key=res.index)
